"Errors that may be returned by Netlab or the Netlab SDK."

from typing import Dict, Optional, Type

from . import klass, lab, pod, reservation, user, vm  # noqa: F401
from .common import NetlabError, ResponseFormatError, UnmappedError


def _construct_mapping() -> Dict[str, Type[NetlabError]]:
    from .common import CODES, NetlabError

    result: Dict[str, Type[NetlabError]] = {}
    for code in CODES:
        assert code.str_error not in result, code.str_error
        result[code.str_error] = code

    return result


NAME_CODE_MAPPING = _construct_mapping()


def error_decode(e):
    if isinstance(e, str):
        error_code = e
        error_data: Optional[str] = None
    elif not isinstance(e, dict):
        return ResponseFormatError()
    else:
        error_code = e["message"]
        try:
            error_data = e["data"]["message"]
        except KeyError:
            error_data = None

    if error_code in NAME_CODE_MAPPING:
        return NAME_CODE_MAPPING[error_code](error_data)
    else:
        return UnmappedError(error_code, error_data)



import asyncio
from pathlib import Path

import click


async def _inner(manifest_paths, auto_confirm):
    import netlab.builder

    await netlab.builder.main(manifest_paths, auto_confirm)


@click.command()
# @click.option("--system", default="default", help="NETLAB+ system to communicate with.")
# @click.option("--jobs", default=1, type=click.IntRange(min=1), help="Number of jobs to run in parallel.")
# @click.option("--jobs_spacing", default=0, type=click.FloatRange(), help="")
@click.option("--yes", "-y", type=bool, is_flag=True)
@click.argument("manifest_paths", type=click.Path(exists=True, readable=True, dir_okay=False, path_type=Path), nargs=1)
@click.pass_context
def build(ctx, manifest_paths: Path, yes):
    # click.confirm("Continue?")
    asyncio.run(_inner([manifest_paths], auto_confirm=yes))



from datetime import datetime
from typing import Any, Dict, List, Optional

from ..utils import minimum_version
from ._client_protocol import ClientProtocol
from .system_results import SystemStatusGetResult


class SystemApiMixin(ClientProtocol):
    async def system_status_get(self, **kwargs) -> SystemStatusGetResult:
        """
        This method allows you to query NETLAB+ system information.

        :return: Returns a system status object with the following properties:

        **Properties**

        cpu_n
            Number of CPUs detected by the virtual appliance.
        uptime_sec
            System uptime in seconds.
        hostname
            NETLAB+ domain name
        sys_lic_exp_date
            License expiration date
        sys_lic_op_state
            License operation stsate
        sys_logins_enabled
            User logins status
        sys_maint_ends
            Mainenance end date
        sys_mode
            System mode
        sys_name
            System name
        sys_product_id
            Product identifier
        sys_sdn_release_date
            Software distribution date
        sys_sdn_release_type
            Software distribution release type
        sys_sdn_version
            Software distribution version
        sys_serial
            System serial number
        """
        method = "system.status.get"
        return await self.call(method, **kwargs)

    @minimum_version("17.1.3")
    async def system_usage_cpu(self, **kwargs):
        """
        This method allows you to query NETLAB+ CPU usage.

        :return: Returns an object with N+1 properties, where N is the number of CPUs. The CPUs are listed from 0 to
                 N-1 and also includes **all** . They each contain an object with the following properties:
        :rtype: dict

        **Properties**

        gnice_pct
            The percentage of time spent by the CPU or CPUs to run a niced guest.
        guest_pct
            The percentage of time spent by the CPU or CPUs to run a virtual processor.
        idle_pct
            The percentage of time that the CPU or CPUs were idle and the system did not have an outstanding disk I/O
            request.
        iowait_pct
            The percentage of time that the CPU or CPUs were idle during which the system had an outstanding disk I/O
            request.
        irq_pct
            The percentage of time spent by the CPU or CPUs to service hardware interrupts.
        nice_pct
            The percentage of CPU utilization that occurred while executing at the user level with nice priority.
        soft_pct
            The percentage of time spent by the CPU or CPUs to service software interrupts.
        steal_pct
            The percentage of time spent in involuntary wait by the virtual CPU or CPUs while the hypervisor was
            servicing another virtual processor.
        sys_pct
            The percentage of CPU utilization that occurred while executing at the system level (kernel). Note that
            this does not include time spent servicing hardware and software interrupts.
        usr_pct
            The percentage of CPU utilization that occurred while executing at the user level (application).
        """
        method = "system.usage.cpu"
        return await self.call(method, **kwargs)

    @minimum_version("17.1.3")
    async def system_usage_disk(self, **kwargs) -> Dict[str, Any]:
        """
        This method allows you to query NETLAB+ disk usage.

        On success, returns an object with multiple properties corresponding to
        the mounted disks in the NETLAB+ appliance. They each contain an object
        with the following properties:

        **Properties**

            avail_b
                         Available disk space in bytes.
            dev_path
                The name of the mounted storage device.
            free_alarm
                Triggered when a disk's free space has dropped below a determined threshold.
            mount_point
                   The directory where the filesystem is mounted.
            netlab_label
                Indicates the purpose of the disk for NETLAB+.
            os_label
                Indicates the purpose of the disk inside the operating system.
            total_b
                Total disk space in bytes.
            used_b
                Used disk space in bytes.
            used_pct
                Disk space used as a percentage of the whole.
            vm_label
                The name of the disk inside the appliance's virtual machine configuration.

        **Warning: Backward Incompatible Change**

        Previous versions of this method returned two properties, data and
        programs . If you were directly referencing one of these two properties
        you may now find them under new names. Refer to the netlab_label property
        in each object to find the objects with the values of Data or Programs .
        """
        method = "system.usage.disk"
        return await self.call(method, **kwargs)

    @minimum_version("17.1.3")
    async def system_usage_memory(self, **kwargs) -> Dict[str, Any]:
        """
        This method allows you to query NETLAB+ memory usage.

        :return: Returns an object with a single property, **ram**. It contains an object with the following properties:

        **Properties**

        free_b
            Free memory in bytes.
        total_b
            Total memory in bytes.
        used_b
            Used memory in bytes.
        """
        method = "system.usage.memory"
        return await self.call(method, **kwargs)

    @minimum_version("17.1.4")
    async def system_time_timezone_get(self, *, tz_id, **kwargs) -> Dict[str, Any]:
        """
        Match a netlab timezone id to the timezone iana.

        :return: Returns an object that represents a timezone.

        **Properties**

        tz_id
            The id of timezone (specific to netlab).
        tz_iana
            The iana identifier of the timezone.
        tz_name
            The name of the timezone.
        tz_sort
            The sort order of the timezone.
        """
        method = "system.time.timezone.get"
        return await self.call(method, tz_id=tz_id, **kwargs)

    @minimum_version("17.1.4")
    async def system_time_timezone_list(self, **kwargs) -> List[Dict[str, Any]]:
        """
        Match a netlab timezone id to the timezone iana.

        :return: Returns an object that represents a timezone.
        :rtype: dict

        **Properties**

        tz_id
            The id of timezone (specific to netlab).
        tz_iana
            The iana identifier of the timezone.
        tz_name
            The name of the timezone.
        tz_sort
            The sort order of the timezone.
        """

        method = "system.time.timezone.list"
        return await self.call(method, **kwargs)

    @minimum_version("19.0.1")
    async def system_perf_query(
        self,
        *,
        start_time: datetime,
        end_time: Optional[datetime] = None,
        metrics: Optional[List[str]] = None,
        sources: Optional[List[str]] = None,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        This method is used to query system performance metrics.

        :param start_time: If specified, only return metrics recorded on or after start time.
        :param end_time: If specified, only return metrics recorded on or before end time.
        :param metrics: If specified, only returns metrics from :ref:`this list<System Performance Metrics>`.
        :param sources: If specified, only returns metrics originating from sources in this list.
            Valid sources include mbusd, podspd, sysspd, respd, userspd, vmspd, histpd, filespd.
        :return: A list of dicts with the following properties.

        **Properties**

        sp_time
            The date/time the metric was recorded.
        sp_metric
            The metric identifier.
        sp_source
            The module/program that produced the metric.
        sp_value
            The numeric value of the metric.
        """

        # Technically start_time is optional, but omitting it causes too much load on the server and
        # returns weird results.

        #  If both start_time and end_time
        # are omitted, the latest observation for each requested metric or source are returned.  If
        # all parameters are omitted, the latest recorded value (if any) for all metrics (if any) are
        # returned.

        method = "system.perf.query"

        params = {"start_time": start_time, "end_time": end_time, "metrics": metrics, "sources": sources}

        return await self.call(method, **params, **kwargs)



from .common import NetlabError, _add


@_add
class AlreadyRegisteredError(NetlabError):
    """Content is already registered to class."""

    str_error = "E_ALREADY_REGISTERED"


@_add
class ClassNotFoundError(NetlabError):
    """Requested cls_id was not found."""

    str_error = "E_CLASS_NOT_FOUND"


@_add
class NotRegisteredError(NetlabError):
    """con_id not registered in class."""

    str_error = "E_NOT_REGISTERED"


@_add
class AccountNotInstructorError(NetlabError):
    """acc_id cannot be set to lead."""

    str_error = "E_ACCOUNT_NOT_INSTRUCTOR"


@_add
class AlreadyMemberError(NetlabError):
    """acc_id is already a member of class."""

    str_error = "E_ALREADY_MEMBER"


@_add
class NotMemberError(NetlabError):
    """acc_id is not on roster."""

    str_error = "E_NOT_MEMBER"


@_add
class RemoveLastLeadError(NetlabError):
    """Returned more than one lead."""

    str_error = "E_REMOVE_LAST_LEAD"



import asyncio
import csv
import io
import sys

import click

from ..async_client import NetlabClient
from ..constants import NL_OK
from ..errors.common import InvalidVersion, NetlabError
from ..serializer import ENUMS, serialize
from .utils import output, print_exception, resolve_enum


@click.command()
@click.pass_context
@click.argument("method", nargs=1)
@click.argument("arguments", nargs=-1)
@click.option("--system", default="default", help="NETLAB+ system to communicate with.")
@click.option("--output", "output_type", default="table", type=click.Choice(["table", "csv", "tsv", "json", "var"]))
def api(ctx, method, arguments, system, output_type):
    """Communicate with a remote NETLAB+ server API"""
    return asyncio.run(_async_api(ctx, method, arguments, system, output_type))


async def _async_api(ctx, method, arguments, system, output_type):
    try:
        async with NetlabClient(system) as netlab_client:
            api_method = getattr(netlab_client, method, None)
            if api_method is None:
                click.echo(
                    "{} is not a valid API method. Please check the documentation.".format(
                        click.style(method, fg="red")
                    )
                )
                ctx.exit(1)
            assert api_method is not None

            api_arguments = {}
            try:
                for arg in arguments:
                    key, value = arg.split("=")
                    key = key.lower()
                    if key in ENUMS:
                        enum = ENUMS[key]
                        value = resolve_enum(enum, value, param=key)
                    api_arguments[key.lower()] = value
            except ValueError:
                # user probably didn't pass arguments in the key=value format
                click.secho("There is an error with your method arguments.", fg="red")
                click.secho("Please ensure they are in the format of arg=value. \n\nExample:")
                click.secho("    {} user_account_get acc_id=100001".format(ctx.command_path), fg="cyan")
                ctx.exit(1)

            try:
                results = await api_method(**api_arguments)
            except (TypeError, NetlabError, InvalidVersion) as err:
                results = None
                click.secho("{}: {}".format(type(err).__name__, err), fg="red")
                click.echo("Please check the documentation for help.")
                ctx.exit(1)

            if results == NL_OK:
                click.secho("Success!", fg="green")
            elif isinstance(results, int) or isinstance(results, str):
                click.echo(results)
            else:
                outputs = {
                    "table": table_output,
                    "csv": csv_output,
                    "tsv": tsv_output,
                    "json": json_output,
                    "var": var_output,
                }
                outputs[output_type](results)  # type: ignore

    except Exception as e:
        print_exception(e)
        sys.exit(1)


def table_output(results):
    if isinstance(results, dict):
        results = [results]
    output(results, headers="keys")


def csv_output(results, delimiter=","):
    out = io.StringIO()

    if isinstance(results, dict):
        results = [results]

    fieldnames = list(results[0].keys())

    writer = csv.DictWriter(out, fieldnames=fieldnames, quoting=csv.QUOTE_MINIMAL, delimiter=delimiter)
    writer.writeheader()
    for result in results:
        writer.writerow(result)
    click.echo(out.getvalue())


def tsv_output(results):
    csv_output(results, delimiter="\t")


def json_output(results):
    click.echo(serialize(results, pretty=True))


def var_output(results):
    if isinstance(results, dict):
        results = [results]

    out = []
    for result in results:
        row = ""
        for key, value in result.items():
            row += "{}={} ".format(key, value)
        out.append(row)
    click.echo("\n".join(out))



from datetime import date
from typing import Any, TypedDict


class SystemStatusGetResult(TypedDict):
    " "

    cpu_n: int
    "Number of CPUs detected by the virtual appliance."
    uptime_sec: int
    "System uptime in seconds."
    hostname: str
    "NETLAB+ domain name"
    sys_lic_exp_date: date
    "License expiration date"
    sys_lic_op_state: bool
    "License operation state"
    sys_logins_enabled: bool
    "User logins status"
    sys_maint_ends: date
    "Mainenance end date"
    sys_mode: Any
    "System mode"
    sys_name: str
    "System name"
    sys_product_id: str
    "Product identifier"
    sys_sdn_release_date: date
    "Software distribution date"
    sys_sdn_release_type: Any
    "Software distribution release type"
    sys_sdn_version: str
    "Software distribution version"
    sys_serial: str
    "System serial number"



"""
Additional Netlab datatypes.
"""

from .common import HDREvent, HDRResult
from .pod import PCCloneSpec
from .vm import VMAlloc, VMAllocHostInfo, VMAllocTotal

__all__ = (
    "VMAlloc",
    "VMAllocTotal",
    "VMAllocHostInfo",
    "PCCloneSpec",
    "HDREvent",
    "HDRResult",
)



from enum import Enum


class ClassLabLimit(Enum):
    """
    Lab limit settings.
    """

    ENFORCE = "E"
    "Enforce class lab limit."
    INSTRUCTOR = "I"
    "Instructor override."


class ClassEmailLogs(Enum):
    """
    Email reservation logs for real equipment pods.
    """

    NO = "N"
    "Don't email logs."
    INSTRUCTOR = "P"
    "Email to primary instructor."


class ClassExtensionSlots(Enum):
    """
    The number of times a reservation can be extended.
    """

    COMMUNITY = -1
    "Use the community setting."


class ContentAccessiblity(Enum):
    """
    Content accessiblity.
    """

    GLOBAL = "G"
    "Globally accessible."
    PRIVATE = "P"
    "Private."



from datetime import datetime
from typing import Any, List, TypedDict

from ..enums import HDRSeverity, HDRStatus


class HDREvent(TypedDict):
    time: datetime
    msg: str
    code: str
    severity: HDRSeverity


class HDRResult(TypedDict):
    status: HDRStatus
    value: Any
    events: List[HDREvent]



import asyncio
import json
import sys

from paepy.ChannelDefinition import CustomSensorResult  # type: ignore

from netlab.async_client import NetlabClient


async def main():
    # interpret first command line parameter as json object
    if len(sys.argv) == 1:
        # for command line debugging, use the default system
        sys_args = '{"params": "default"}'
    else:
        sys_args = sys.argv[1]
    prtg = json.loads(sys_args)
    system = prtg.get("params", "default").lower()

    # PRTG runs sensors as a different user. Specify the Administrator user's netlab config file here
    async with NetlabClient(system, config_path="C:/Users/Administrator/.netlab/config.json") as client:
        # query data from netlab api
        response = await client.system_usage_disk()
        data_free_space = int(100 - response["data"]["used_pct"])
        data_free_bytes = response["data"]["total_b"] - response["data"]["used_b"]
        programs_free_space = int(100 - response["programs"]["used_pct"])
        programs_free_bytes = response["programs"]["total_b"] - response["programs"]["used_b"]

        # create sensor result
        result = CustomSensorResult("OK")

        # add primary channel for percentage of data partition free space
        result.add_channel(
            channel_name="Free Space - data",
            primary_channel=True,
            unit="Percent",
            value=data_free_space,
            is_limit_mode=True,
            limit_min_warning=25,
            limit_min_error=10,
        )

        # additional channels for remaining values
        result.add_channel(channel_name="Free Bytes - data", unit="BytesDisk", value=data_free_bytes)
        result.add_channel(channel_name="Free Space - programs", unit="Percent", value=programs_free_space)
        result.add_channel(channel_name="Free Bytes - programs", unit="BytesDisk", value=programs_free_bytes)

        # print sensor result to stdout
        print(result.get_json_result())


if __name__ == "__main__":
    asyncio.run(main())



import json
import logging
import os
import ssl
import warnings
from ipaddress import ip_address
from pathlib import Path
from typing import Any, Dict, List, Literal, NamedTuple, Optional, Union

from .config import NetlabCipherListEnum
from .errors.common import InvalidConfig

ENV_PREFIX = "NETLAB_CONFIG_"
ENV_BOOLEANS: List[str] = []
CONFIG_FILENAME = os.path.join(".netlab", "config.json")
DEFAULT_CONFIG_PATH = os.path.join(os.path.expanduser("~"), CONFIG_FILENAME)

log = logging.getLogger()


class SystemConfig(NamedTuple):
    user: str
    token: str
    host: str
    port: int
    ssl: Union[ssl.SSLContext, Literal["default", "noverify"]]
    ssl_ciphers: Optional[str]
    server_hostname: Optional[str]
    message_byte_limit: int


def get_system_config(system, config=None, config_path=None) -> SystemConfig:
    defaults: dict[str, Any] = {
        "host": "localhost",
        "user": "administrator",
        "port": 9000,
        "ssl": "default",
        "server_hostname": None,
        "message_byte_limit": 1024 * 1024 * 64,
        "ssl_ciphers": "NETLAB",
    }

    if config:
        if isinstance(config, str):
            if config[:3] == "py:":
                import importlib

                _, module_spec, attr_spec = config.split(":")
                module = importlib.import_module(module_spec)
                c = module
                for k in attr_spec.split("."):
                    c = getattr(c, k)
                config = c

        log.debug("using explicit config")
        # config from Client init
        defaults.update(config)
    else:
        log.debug("collecting config from system")
        # file config
        add_config = load_config(system, config_path)
        log.debug(
            f"collecting config from filesystem {system=} {config_path=} {config=}",
        )
        defaults.update(add_config)

        # environment variables config
        env_vars = [(k, v) for (k, v) in os.environ.items() if k.startswith(ENV_PREFIX)]
        if env_vars:
            env_config: Dict[str, Union[str, bool]] = {}
            for k, v in env_vars:
                key = k.replace(ENV_PREFIX, "").lower()
                if key in ENV_BOOLEANS:
                    # convert env value to python boolean
                    env_config[key] = v.lower() in ("true", "yes", "t", "1")
                else:
                    env_config[key] = v
            log.debug(
                f"collecting config from envvars {env_config=}",
            )
            defaults.update(env_config)

    try:
        ip = ip_address(defaults["host"])
    except ValueError:
        ip = None

    if defaults["server_hostname"] is None and ip is not None and defaults["ssl"]:
        log.warn("You have passed an ip address without a 'server_hostname'. This is insecure.")
        defaults["server_hostname"] = ""
    if "user" not in defaults:
        raise InvalidConfig('Config missing required "user".')
    if "token" not in defaults:
        raise InvalidConfig('Config missing required "token".')
    if "timeout" in defaults:
        defaults.pop("timeout")
        warnings.warn("Config contains `timeout`, which is deprecated.", DeprecationWarning)
    if not defaults["ssl"]:
        warnings.warn("SSL is not being used, this is insecure.", DeprecationWarning)

    defaults["ssl_ciphers"] = NetlabCipherListEnum.get_cipher_list(defaults["ssl_ciphers"])

    config = SystemConfig(**defaults)

    log.debug(f"collected config {config}")
    return config


def load_config(system=None, config_path=None):
    """
    Loads authentication data from a Netlab configuration file in the given
    root directory or if config_path is passed use given path.
    If no system is specified, return entire config dict.
    Lookup priority:
        explicit config_path parameter > NETLAB_CONFIG environment variable > ~/.netlab/config.json
    """
    config_file = find_config_file(config_path)

    if not config_file or not config_file.exists():
        return {}

    log.debug(f"config file found {config_file=}")
    try:
        with open(config_file) as f:
            data = json.load(f)
            if system:
                return data.get(system, {})
            else:
                return data
    except (IOError, KeyError, ValueError) as e:
        log.debug(e)
        raise


def add_config(system, config, config_path=None):
    """
    Adds system and config data from a Netlab to the configuration file
    """
    current_config = load_config(config_path=config_path)
    current_config[system] = config
    config_file = find_config_file(config_path)
    if not config_file:
        env_path = os.environ.get("NETLAB_CONFIG")
        if env_path:
            config_file = Path(env_path) / "config.json"
        else:
            config_file = Path(DEFAULT_CONFIG_PATH)

    try:
        os.makedirs(os.path.dirname(config_file), exist_ok=True)
        with open(config_file, "w") as f:
            json.dump(current_config, f, sort_keys=True, indent=4)
    except (IOError, KeyError, ValueError) as e:
        log.debug(e)
        raise


def remove_config(system, config_path=None):
    """
    Removes system and config data from the configuration file
    """
    current_config = load_config(config_path=config_path)
    if not system:
        current_config = {}
    else:
        current_config.pop(system, None)

    config_file = find_config_file(config_path)

    try:
        with open(config_file, "w") as f:
            json.dump(current_config, f, sort_keys=True, indent=4)
    except (IOError, KeyError, ValueError) as e:
        log.debug(e)
        raise


def find_config_file(config_path: Optional[str] = None) -> Path:
    netlab_config_path = os.environ.get("NETLAB_CONFIG")

    environment_path = (
        os.path.join(netlab_config_path, os.path.basename(CONFIG_FILENAME)) if netlab_config_path else None
    )

    paths = filter(None, [config_path, environment_path, DEFAULT_CONFIG_PATH])

    return Path(next(paths))



import dataclasses
import logging
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional

from ..async_client import NetlabClient
from ..enums import VirtualMachineRole
from . import pod_op, schema
from .schema import BuilderPCCloneSpec, ThrottlerConfig
from .types import ExistingPod, NetlabHost

log = logging.getLogger(__name__)


def client_factory_from_manifest(manifest: schema.Manifest) -> NetlabClient:
    if manifest.system is None:
        client_fact = NetlabClient()
    elif isinstance(manifest.system, schema.SystemSpec):
        client_fact = NetlabClient(config=manifest.system.config)
    elif isinstance(manifest.system, schema.SystemSpec1):
        client_fact = NetlabClient(system=manifest.system.name)
    else:
        client_fact = NetlabClient(config=manifest.system)

    return client_fact


@dataclass
class CompiledDeletePod:
    pod_id: int
    parent: "CompiledDelete"

    existing_pod: ExistingPod | None = None
    superseded_by: Optional["CompiledPod"] = None

    def get_conflicts(self) -> list[tuple[str, Any, Any]]:
        if self.existing_pod is not None:
            return [("state", "ABSENT", "EXISTS")]
        else:
            return []

    def pod_op(self) -> pod_op.PodOp:
        if self.superseded_by is not None:
            return pod_op.Superceded(pod=self, by=self.superseded_by)
        elif self.existing_pod is not None:
            return pod_op.Delete(pod=self, existing=self.existing_pod)
        else:
            return pod_op.Absent(pod=self)
        # conflicts = pod.get_conflicts()


@dataclass
class CompiledPod:
    pod_id: int
    parent: "CompiledRange"

    existing_pod: ExistingPod | None = None

    def get_host(self) -> NetlabHost | None:
        return self.parent.host

    def get_role(self) -> VirtualMachineRole:
        return self.parent.get_role()

    def get_pod_type(self) -> str:
        assert self.parent.source_pod
        return self.parent.source_pod.get_pod_type()

    def pod_op(self) -> pod_op.PodOp:
        if self.existing_pod is None:
            return pod_op.Create(
                pod=self,
            )
        else:
            conflicts = [
                ("pod_name", self.get_name(), self.existing_pod.get_name()),
                ("pt_id", self.parent.get_pod_type(), self.existing_pod.current_pod_type),
                ("vh_host", self.parent.host, self.existing_pod.current_host),
                ("vm_role", self.parent.get_role(), self.existing_pod.current_role),
            ]
            conflicts = list(filter(lambda x: x[1] != x[2], conflicts))

            if len(conflicts) == 0:
                return pod_op.Present(pod=self, existing=self.existing_pod)
            else:
                return pod_op.Rebuild(
                    pod=self,
                    existing=self.existing_pod,
                    conflicts=conflicts,
                )

    # def get_conflicts(self) -> list[tuple[str, Any, Any]]:
    #     if self.existing_pod is None:
    #         return [("state", "EXISTS", "ABSENT")]
    #
    #     results = [
    #         ("pod_name", self.get_name(), self.existing_pod.get_name()),
    #         ("pt_id", self.parent.get_pod_type(), self.existing_pod.current_pod_type),
    #         ("vh_host", self.parent.host, self.existing_pod.current_host),
    #         ("vm_role", self.parent.get_role(), self.existing_pod.current_role),
    #     ]
    #
    #     return list(filter(lambda x: x[1] != x[2], results))

    def get_name(self) -> str:
        vh = self.parent.host
        assert vh
        vh_id, vh_name = vh
        ip_split = re.split(r"[.-]", vh_name)
        host_suffix = "H" + ip_split[-1]

        match self.parent.get_role():
            case VirtualMachineRole.MASTER:
                pod_kind = "M"
            case VirtualMachineRole.NORMAL | VirtualMachineRole.PERSISTENT:
                pod_kind = "P"
            case VirtualMachineRole.TEMPLATE:
                pod_kind = "T"

        clone_name = f"{host_suffix}.{pod_kind}{self.pod_id}"
        if self.parent.clone_prefix is not None:
            clone_name = self.parent.clone_prefix + "." + clone_name
        if self.parent.clone_suffix is not None:
            clone_name = clone_name + "." + self.parent.clone_suffix

        context = {
            "pod_id": self.pod_id,
            "pod_role": pod_kind,
            "vh_name": ip_split[-1],
            "pt_org": self.parent.get_existing_pod().pt_org,
            "pt_apdid": self.parent.get_existing_pod().pt_apdid,
            "default": clone_name,
        }

        return self.parent.clone_name_template.format(**context)


@dataclass
class CompiledDelete:
    pods: list[CompiledDeletePod]


@dataclass
class CompiledRangeErrors:
    errors: list[tuple[str, str]]


@dataclass
class CompiledRange:
    pod_online: bool
    source_pod: ExistingPod | CompiledPod
    pc_clone_specs: schema.BuilderPCCloneSpec | None
    clone_prefix: str | None
    clone_suffix: str | None
    clone_name_template: str

    pods: list[CompiledPod] = dataclasses.field(default_factory=list)
    host: NetlabHost | None = None

    def get_existing_pod(self) -> ExistingPod:
        match self.source_pod:
            case ExistingPod():
                return self.source_pod
            case CompiledPod():
                return self.source_pod.parent.get_existing_pod()

    def get_pod_type(self) -> str:
        assert self.source_pod is not None
        return self.source_pod.get_pod_type()

    def get_role(self) -> VirtualMachineRole:
        if self.pc_clone_specs is not None:
            clone_role = self.pc_clone_specs.get("clone_role", "NORMAL")
            return VirtualMachineRole[clone_role]
        else:
            return VirtualMachineRole.NORMAL

    # def get_host(self) -> NetlabHost | None:
    #     if self.host is not None:
    #         return self.host
    #     else:
    #         return self.source_pod.get_host()


def _host_from_pod(
    host_map: dict[int | str, NetlabHost], pc_clone_specs: schema.BuilderPCCloneSpec | None
) -> NetlabHost | None:
    if pc_clone_specs is None:
        host = None
        pc_clone_specs = None
    else:
        clone_vh_id = pc_clone_specs.get("clone_vh_id")
        clone_vh_name = pc_clone_specs.get("clone_vh_name")

        match (clone_vh_id, clone_vh_name):
            case (None, None):
                host = None
            case (int(clone_vh_id), None):
                host = host_map[clone_vh_id]
            case (None, str(clone_vh_name)):
                host = host_map[clone_vh_name]
            case (int(clone_vh_id), str(clone_vh_name)):
                host = host_map[clone_vh_name]
                assert host.vh_id == clone_vh_id
                assert host.vh_name == clone_vh_name
    return host


@dataclass
class CompilerError(Exception):
    code: str
    desc: str


@dataclass
class CompiledManifest:
    manifest_path: Path
    client: NetlabClient
    throttle_delete: ThrottlerConfig
    throttle_create: ThrottlerConfig
    throttle_updown: ThrottlerConfig
    ranges: list[CompiledRange | CompiledDelete | CompiledRangeErrors] = dataclasses.field(default_factory=list)

    def _compile_delete_range(
        self,
        existing_pods: dict[int, ExistingPod],
        compiled_pods: dict[int, CompiledPod],
        compiled_delete_pods: dict[int, CompiledDeletePod],
        r,
    ):
        compiled_delete_range = CompiledDelete(pods=[])
        self.ranges.append(compiled_delete_range)
        for p in r.get_range():
            compiled_delete_pod = CompiledDeletePod(
                p,
                existing_pod=existing_pods.get(p),
                parent=compiled_delete_range,
            )
            assert compiled_delete_pod.pod_id not in compiled_pods
            compiled_delete_pod = compiled_delete_pods.setdefault(p, compiled_delete_pod)
            compiled_delete_range.pods.append(compiled_delete_pod)

    def _compile_range(self):
        pass

    @classmethod
    def from_schema(
        cls,
        manifest: schema.Manifest,
        manifest_path: Path,
        host_map: dict[int | str, NetlabHost],
        existing_pods: dict[int, ExistingPod],
    ) -> "CompiledManifest":
        compiled_manifest = CompiledManifest(
            manifest_path=manifest_path,
            client=client_factory_from_manifest(manifest),
            throttle_create=manifest.throttles.create,
            throttle_delete=manifest.throttles.delete,
            throttle_updown=manifest.throttles.updown,
        )
        compiled_pods: dict[int, CompiledPod] = {}
        compiled_delete_pods: dict[int, CompiledDeletePod] = {}

        for r in manifest.range:
            if r.delete:
                compiled_manifest._compile_delete_range(
                    existing_pods,
                    compiled_pods,
                    compiled_delete_pods,
                    r,
                )
                continue
            else:
                pass

            errors = []
            pc_clone_specs: BuilderPCCloneSpec = {}

            source_pod = compiled_pods.get(r.source_pod_id, existing_pods.get(r.source_pod_id))
            if source_pod is None:
                errors.append(("NO_SOURCE_POD", ""))
            else:
                try:
                    if r.pc_clone_specs:
                        pc_clone_specs.update(r.pc_clone_specs)

                    host = _host_from_pod(host_map, pc_clone_specs)
                    if host is None:
                        host = source_pod.get_host()

                    if host is None:
                        errors.append(("NO_CLONE_HOST", "The host to clone to could not be determined."))
                    else:
                        pc_clone_specs["clone_vh_id"] = host.vh_id
                        pc_clone_specs.pop("clone_vh_name", None)  # TODO
                except KeyError as e:
                    errors.append(("HOST_NOT_FOUND", f"Host '{e.args[0]}' not found."))

                if (
                    source_pod.get_role() is not VirtualMachineRole.TEMPLATE
                    and pc_clone_specs
                    and pc_clone_specs.get("clone_type") != "FULL"
                ):
                    errors.append(
                        (
                            "FULL_CLONE_REQUIRED",
                            f"Source pod {source_pod.pod_id} is not a TEMPLATE so a FULL clone is required.",
                        )
                    )

            if errors:
                compiled_manifest.ranges.append(CompiledRangeErrors(errors=errors))
                continue
            else:
                assert source_pod is not None

            compiled_range = CompiledRange(
                pod_online=r.pod_online,
                pc_clone_specs=pc_clone_specs,
                host=host,
                clone_prefix=r.clone_prefix if r.clone_prefix is not False else None,
                clone_suffix=r.clone_suffix if r.clone_suffix is not False else None,
                clone_name_template=r.clone_name_template,
                source_pod=source_pod,
            )
            compiled_manifest.ranges.append(compiled_range)

            for p in r.get_range():
                compiled_pod = CompiledPod(
                    p,
                    existing_pod=existing_pods.get(p),
                    parent=compiled_range,
                )
                compiled_pod = compiled_pods.setdefault(p, compiled_pod)
                if compiled_pod.pod_id in compiled_delete_pods:
                    compiled_delete_pods[compiled_pod.pod_id].superseded_by = compiled_pod
                compiled_range.pods.append(compiled_pod)

        return compiled_manifest



from enum import Enum


class ReservationType(Enum):
    "Type of reservation."

    INDIVIDUAL = "S"  # 'INDIVIDUAL'
    "Individual student reservation."
    ILT_CLASS = "C"  # 'ILT_CLASS'
    "Instructor-led class reservation."
    TEAM = "T"  # 'TEAM'
    "Team student reservation."
    INSTRUCTOR = "I"  # 'INSTRUCTOR'
    "Instructor reservation for own use."


class ReservationScope(Enum):
    "Reservation scope."

    RELEVANT = "RELEVANT"
    "Relevant to caller, i.e. reservations associated with you in context to user role."
    ALL = "ALL"
    "All reservations."
    COMMUNITY = "COMMUNITY"
    "Relevant to the community."
    CLASS = "CLASS"
    "Relevant to your class."


class InitConfig(Enum):
    "Initial configuration."

    NONE = "N"
    "No configs loaded."
    LAST = "L"
    "Load last saved config (if any)."
    EXERCISE = "E"
    "Load default for exercise (if any)."



import json
from datetime import date, datetime, timedelta
from decimal import Decimal
from enum import Enum
from typing import Any
from uuid import UUID

from . import enums
from .errors import ResponseFormatError

ENUMS = {
    "cls_email_logs": enums.ClassEmailLogs,
    "cls_lab_limit": enums.ClassLabLimit,
    "acc_type": enums.AccountType,
    "pod_cat": enums.PodCategory,
    "pod_admin_state": enums.PodAdminState,
    "pod_current_state": enums.PodCurrentState,
    "pc_icon": enums.PCIcon,
    "pc_type": enums.PCType,
    "pl_icon": enums.PLIcon,
    "vh_com_path": enums.VirtualHostComPath,
    "vm_role": enums.VirtualMachineRole,
    "res_type": enums.ReservationType,
    "date_format": enums.DateFormat,
    "time_format": enums.TimeFormat,
    "cls_ext_slots_per_res": (enums.ClassExtensionSlots, int),
    "com_ext_slots_per_res": (enums.CommunityExtensionSlots, int),
    "con_access": enums.ContentAccessiblity,
    "free_alarm": enums.FreeAlarm,
}

ENUM_CSV = {
    "acc_privs": enums.AccountPrivileges,
    # Pod
    "pod_cat_values": enums.PodCategory,
}

STR_CSV = {
    "con_pt_ids",
}

DATE_FIELDS = {
    # Class
    "cls_end_date",
    "cls_start_date",
    # System
    "sys_lic_exp_date",
    "sys_maint_ends",
    "sys_sdn_release_date",
    # Reservation
    "local_ymd",
    # History
    "crs_release_date",
}

DATETIME_FIELDS = {
    # HDR
    "time",
    "hdr_time",
    # User
    "acc_last_login",
    "acc_session_time_started",
    "acc_time_created",
    "acc_time_last_access",
    # Pod
    "time",
    "vh_bios_date",
    "vh_date_modified",
    "vh_date_tested",
    "vm_date_added",
    "vh_date_added",
    "vm_date_added",
    # Reservation
    "res_active_time",
    "res_attend_time",
    "res_end",
    "res_post_time",
    "res_preload_time",
    "res_start",
    "utc",
    "local",
    "time_in_local",
    "time_in_utc",
    "time_out_local",
    "time_out_utc",
    "res_end_utc",
    "res_start_utc",
    "start_time",
    "end_time",
    # History
    "lh_active_time",
    "lh_attend_time",
    "lh_end_time",
    "lh_start_time",
    "res_active_time",
    "res_attend_time",
    "res_end_actual",
    "res_end_sched",
    "res_load_time",
    "res_post_time",
    "res_save_time",
    "res_start",
    "res_term_time",
    "rh_report_time"
    # VM
    "vdc_date_added",
    "vdc_date_tested",
    "vh_bios_date",
    "vh_date_added",
    "vh_date_modified",
    "vh_date_tested",
    "vm_date_added",
    "vhp_time",
    # System
    "sp_time",
}

TIMEDELTA_FIELDS = {
    # Reservation
    "res_remaining_dhms",
}

BOOLEAN_FIELDS = {
    # Class
    "cls_change_ex",
    "cls_no_delete",
    "cls_retain_ilt",
    "cls_retain_st",
    "cls_self_sched",
    "cls_team_sched",
    "lead",
    "con_managed",
    "con_global",
    "registered",
    "auth_all_communities",
    # System
    "sys_logins_enabled",
    # User
    "acc_can_login",
    "acc_pw_change",
    "acc_sys",
    "test_taken",
    "com_enabled",
    "enabled",
    "logout",
    "changed",
    "acc_nlx_terms",
    "acc_npd_terms",
    # Pod
    "pod_acl_enabled",
    "pod_auto_net_enabled",
    "pod_auto_net_host_setup",
    "pod_auto_net_host_teardown",
    "pod_managed",
    "pt_removable",
    "pc_online",
    "pc_revert_to_scrub",
    "pc_vnc_use_copy_rect",
    "vh_online",
    "vh_pra_enabled",
    "vm_auto_display",
    "vm_auto_network",
    "vm_auto_settings",
    "vm_sanity_checks",
    "pc_revert_to_scrub",
    "pc_vnc_use_copy_rect",
    "vh_pra_enabled",
    "vm_auto_display",
    "vm_auto_network",
    "vm_auto_settings",
    "vm_sanity_checks",
    "pt_removable",
    "pod_managed",
    "pod_dyn_vlan"
    # Reservation
    "res_done",
    "res_is_active",
    "tz_is_dst",
    # History
    "res_done",
    "rh_reported"
    # VM
    "notify_progress",
    "notify_complete",
    "vh_online",
    "vh_pra_enabled",
    "data_available"
    # Paging
    "out_of_bounds",
}

DECIMAL_FIELDS = {
    # System
    "gnice_pct",
    "guest_pct",
    "idle_pct",
    "iowait_pct",
    "irq_pct",
    "nice_pct",
    "soft_pct",
    "steal_pct",
    "sys_pct",
    "usr_pct",
    "uptime_sec",
    "used_pct",
    "sp_value",
}

INT_FIELDS = {
    # Class
    "cls_max_slots_per_res",
    "cls_min_hours_btw_res",
    "cls_retain_period",
    "com_idcls_id",
    "acc_id",
    "con_trustee",
    # Lab
    "con_build",
    "con_fs_id",
    "con_trustee",
    "ex_index",
    "ex_minutes",
    "ex_fs_id",
    # Pod
    "acc_id",
    "pod_id",
    "pod_csw_base_port",
    "pod_csw_base_vlan",
    "pod_csw_id",
    "pod_id",
    "pod_id_avail_count",
    "pod_id_base",
    "pod_id_range_min",
    "pod_id_range_max",
    "pod_osm_enabled",
    "non_osm_pod_count",
    "pod_vlan_pool_size",
    "sys_total_pod_count",
    "clone_vh_id",
    "pl_index",
    "dev_id",
    "line",
    "pc_id",
    "pev_id",
    "pod_id",
    "res_id",
    "severity",
    "device_count",
    "pt_apc_port_count",
    "pt_as_port_count",
    "pt_build",
    "pt_csw_port_count",
    "pt_pod_max",
    "pt_vlan_pool",
    "remote_pc_count",
    "pod_res_id",
    "pc_pod_id",
    "pc_pod_index",
    "pc_vnc_port",
    "pl_index",
    "vdc_id",
    "vh_cpu_cores",
    "vh_cpu_mhz",
    "vh_cpu_threads",
    "vh_cpu_threads",
    "vh_id",
    "vh_memory_mb",
    "vh_pra_max_cpu",
    "vh_pra_max_mem_mb",
    "vh_pra_max_vm",
    "vhg_id",
    "vm_alloc_cpu_n",
    "vm_alloc_mem_mb",
    "vm_child_count",
    "vm_id",
    "vm_parent_id",
    "vm_runtime_vh_id",
    "pt_apc_port_count",
    "pt_as_port_count",
    "pt_build",
    "pt_csw_port_count",
    "pt_index",
    "pt_pod_max",
    "pt_vlan_pool",
    "remote_pc_count",
    "pt_apc_port_count",
    "pt_as_port_count",
    "pt_csw_port_count",
    "pt_index",
    "pt_pod_max",
    "pt_vlan_pool",
    "remote_pc_count",
    # Reservation
    "acc_com_id",
    "acc_id",
    "cls_div_id",
    "cls_id",
    "pod_id",
    "res_acc_id",
    "res_cls_id",
    "res_com_id",
    "res_flags",
    "res_id",
    "res_init_fail",
    "res_minutes",
    "res_pod_id",
    "res_remaining_sec",
    "res_id",
    "res_minutes",
    "serial",
    "slot_minutes",
    "acc_com_id",
    "cls_div_id",
    "cls_id",
    "pod_id",
    "res_acc_id",
    "res_cls_id",
    "res_com_id",
    "res_flags",
    "res_init_fail",
    "res_minutes",
    "res_pod_id",
    "active_reservations",
    "completed_reservations",
    "future_reservations",
    "pods_in_use",
    "total_reservations",
    "days",
    "hours",
    "minutes",
    "months",
    "local_day",
    "local_day_of_week",
    "local_hour_12",
    "local_hour_24",
    "local_minute",
    "local_second",
    "local_year",
    "tz_id",
    "tz_id_in",
    "tz_id_out",
    # History
    "acc_id",
    "cli_bytes_in",
    "cli_bytes_out",
    "cli_connects",
    "cli_seconds",
    "cls_id",
    "com_id",
    "lh_attend_min",
    "lh_n_users",
    "vnc_bytes_in",
    "vnc_bytes_out",
    "vnc_connects",
    "vnc_seconds",
    "pod_id",
    "res_extensions",
    "res_id",
    "rh_attend_min",
    "rh_init_fail_total",
    "rh_n_lab_devices",
    "rh_n_users",
    "rh_n_vcpu",
    "rh_n_vm",
    "rh_n_vm_hosts",
    "rh_resume_total",
    "rh_suspend_total"
    # System
    "cpu_n",
    "free_b",
    "avail_b",
    "total_b",
    "used_b",
    # User
    "acc_id",
    "acc_fs_id",
    "acc_logins",
    "com_id",
    "tz_id",
    "com_alt_banner_height",
    "com_alt_banner_width",
    "com_max_slots_per_res",
    "com_min_hours_btw_res",
    "active",
    "logouts",
    "acc_fs_id",
    "acc_maint_notify",
    "acc_session_tx_number",
    "acc_sys",
    "acc_session_tx_number",
    "accman_list_com_id",
    "first_weekday",
    # VM
    "progress",
    "vdc_id",
    "vh_id",
    "vh_cpu_cores",
    "vh_cpu_count",
    "vh_cpu_mhz",
    "vh_cpu_threads",
    "vh_memory_mb",
    "vh_pra_max_cpu",
    "vh_pra_max_mem_mb",
    "vh_pra_max_vm",
    "vm_id",
    "pc_id",
    "pc_pod_id",
    "pc_vm_id",
    "vdc_id",
    "vh_id",
    "vhg_id",
    "vm_alloc_cpu_n",
    "vm_alloc_mem_mb",
    "vm_child_count",
    "vm_parent_id",
    "age_sec",
    # Paging
    "page_length",
    "total_records",
    "total_pages",
    "page_limit",
    "current_page",
    "page_offset",
    "page_total",
}

UUID_FIELDS = {
    # History
    "acc_uuid",
    "cls_uuid",
    "com_uuid",
    "crs_uuid",
    "lh_uuid",
    "rh_uuid",
    "lh_uuid_current",
    "pod_uuid",
    "res_uuid",
    "rh_uuid",
}

BLANK_IS_NONE_FIELDS = {
    # class
    "ros_team",
}


def deserialize(data):
    return json.loads(data, object_hook=process)


def resolve_enum(enum, value):
    if not isinstance(enum, tuple):
        enum = (enum,)
    for e in enum:
        try:
            return e(value)
        except:  # noqa
            pass
    raise ResponseFormatError("The value {} could not resolve to one of '{}'.".format(value, enum))


def process(obj):
    for field, value in obj.items():
        if field in DATE_FIELDS:
            obj[field] = to_date(value)
        if field in DATETIME_FIELDS:
            obj[field] = to_datetime(value)
        if field in TIMEDELTA_FIELDS:
            obj[field] = to_timedelta(value)
        if field in BOOLEAN_FIELDS:
            obj[field] = to_bool(value)
        if field in DECIMAL_FIELDS:
            obj[field] = to_decimal(value)
        if field in INT_FIELDS:
            obj[field] = to_int(value)
        if field in UUID_FIELDS:
            obj[field] = to_uuid(value)
        if field in ENUMS and value is not None:
            obj[field] = resolve_enum(ENUMS[field], value)
        if field in ENUM_CSV and value is not None:
            enum = ENUM_CSV[field]
            obj[field] = list(map(lambda v: enum[v], value.split(",")))
        if field in STR_CSV:
            if isinstance(value, str):
                obj[field] = value.split(",")
        if field in BLANK_IS_NONE_FIELDS and value == "":
            obj[field] = None
    return obj


def to_date(value):
    if value:
        return datetime.strptime(value, "%Y-%m-%d").date()
    else:
        return None


def to_datetime(value):
    if value:
        try:
            return datetime.strptime(value, "%Y-%m-%d %H:%M:%S")
        except ValueError:
            # rarely timestamps include microseconds. catch and process those
            return datetime.strptime(value, "%Y-%m-%d %H:%M:%S.%f")
    else:
        return None


def to_timedelta(value):
    if value:
        days, hours, minutes, seconds = [int(i) for i in value.split(",")]
        return timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)
    else:
        return None


def to_bool(value):
    truth_values = ("1", "true", "yes")
    normalized_value = str(value).lower()
    return normalized_value in truth_values


def to_decimal(value):
    if value:
        return Decimal(value)
    else:
        return None


def to_int(value):
    if value:
        return int(value)
    else:
        return 0


def to_uuid(value):
    if value:
        return UUID(value)
    else:
        return None


def serialize(data, pretty=False):
    extra_args: Any = {}
    if pretty:
        extra_args["indent"] = 2
        extra_args["sort_keys"] = True
    return json.dumps(data, default=default, **extra_args)


def default(obj):
    if isinstance(obj, Enum):
        return obj.value
    if isinstance(obj, datetime):
        return from_datetime(obj)
    if isinstance(obj, date):
        return from_date(obj)
    if isinstance(obj, timedelta):
        return from_timedelta(obj)
    if isinstance(obj, UUID):
        return from_uuid(obj)
    if isinstance(obj, Decimal):
        return from_decimal(obj)
    return obj


def from_date(value):
    return value.strftime("%Y-%m-%d")


def from_datetime(value):
    return value.strftime("%Y-%m-%d %H:%M:%S")


def from_timedelta(value):
    days, secs = value.days, value.seconds
    hours = secs // 3600
    minutes = (secs % 3600) // 60
    seconds = secs % 60
    return ",".join([str(days), str(hours), str(minutes), str(seconds)])


def from_uuid(value):
    return str(value)


def from_decimal(value):
    return str(value)



import asyncio
from pprint import pprint

from netlab.async_client import NetlabClient


async def main():
    async with NetlabClient() as connection:
        info = await connection.system_status_get()
        pprint(info)


if __name__ == "__main__":
    asyncio.run(main())



# flake8: noqa
import asyncio
from netlab.async_client import NetlabClient


# Create a function that we can use `await` in
async def main():
    # Create and connect the NetlabConnection
    async with NetlabClient() as client:
        # List communities
        result = await client.user_community_list()
        print(">>>>", result)
        # List used pod ids
        result = await client.pod_list_used_ids()
        print(">>>>", result)

        # Iterate over user accounts
        async for user in client.user_account_search_iter(properties=["acc_id"]):
            print(user)

        # Make 10 requests simultaneously
        result = await asyncio.gather(*[client.user_account_search(properties=["acc_id"], page=1) for i in range(10)])
        print(result)


if __name__ == "__main__":
    asyncio.run(main())



from typing import Annotated as A
from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field
from pydantic.types import StringConstraints
from typing_extensions import NotRequired, Self

from netlab.throttle import Throttler

from ..datatypes.pod import PCCloneSpec


class BuilderPCCloneSpec(PCCloneSpec):
    clone_vh_name: NotRequired[str]
    "The runtime host of cloned VM. ()"


class Spec(BaseModel):
    model_config = ConfigDict(extra="forbid")


class ThrottlerConfig(Spec):
    buckets: int = 1
    refresh: float = 1.0
    stagger: float = 0.0

    def make(self) -> Throttler:
        return Throttler(
            buckets=self.buckets,
            refresh=self.refresh,
            stagger=self.stagger,
        )


# class CloneSpecs(Spec):
#     clone_datastore: str
#     clone_role: str = "NORMAL"
#     clone_type: str = "LINKED"


class DeploymentThrottlerConfigs(Spec):
    EXPERIMENTAL_I_know_what_I_am_doing: Literal[True]

    delete: ThrottlerConfig = Field(default_factory=lambda: ThrottlerConfig(buckets=5, refresh=0, stagger=0.5))
    create: ThrottlerConfig = Field(default_factory=lambda: ThrottlerConfig(buckets=1, refresh=0, stagger=1))
    updown: ThrottlerConfig = Field(default_factory=lambda: ThrottlerConfig(buckets=20, refresh=0, stagger=0))
    fetch: ThrottlerConfig = Field(default_factory=lambda: ThrottlerConfig(buckets=20, refresh=0, stagger=0))

    @classmethod
    def default(cls) -> Self:
        return cls(EXPERIMENTAL_I_know_what_I_am_doing=True)


class DeploymentPlan(Spec):
    # vh_uuid: UUID

    pod_online: bool = False

    delete: bool = False

    source_pod_id: int

    start: int
    number: int
    step: int = 1

    clone_prefix: A[str, StringConstraints(min_length=1)] | Literal[False] = False
    clone_suffix: A[str, StringConstraints(min_length=1)] | Literal[False] = False
    clone_name_template: str = "{default}"

    pc_clone_specs: BuilderPCCloneSpec | None = None

    def get_range(self) -> range:
        range_to = self.start + self.number * self.step
        return range(self.start, range_to, self.step)


class SystemSpec(Spec):
    config: str


class SystemSpec1(Spec):
    name: str


class Manifest(Spec):
    system: SystemSpec | SystemSpec1 | dict[str, Any] | None = None
    range: list[DeploymentPlan]
    throttles: DeploymentThrottlerConfigs = Field(default_factory=DeploymentThrottlerConfigs.default)



import asyncio
import copy
import logging
import logging.config
import shutil
import sys
import tempfile
from pathlib import Path
from typing import Any

import click

from .. import debugger

LOGGING = {
    "version": 1,
    "root": {
        "handlers": ["rich"],
        "level": "DEBUG",
    },
    "formatters": {
        "brief": {"format": "%(message)s"},
        "default": {"format": "%(asctime)s %(levelname)-8s %(module)-15s %(message)s"},
    },
    "handlers": {
        "rich": {"class": "rich.logging.RichHandler", "formatter": "brief"},
    },
}


log = logging.getLogger()


async def run_install_wormhole():
    proc = await asyncio.create_subprocess_exec(sys.executable, "-m", "pip", "install", "magic-wormhole")

    await proc.communicate()


async def run_wormhole(wormhole_cmd, file_name):
    proc = await asyncio.create_subprocess_exec(
        wormhole_cmd,
        "send",
        file_name,
    )

    await proc.communicate()


def find_wormhole():
    wormhole_cmd = shutil.which("wormhole")
    if wormhole_cmd is not None:
        return wormhole_cmd
    wormhole_cmd = shutil.which(str(Path(sys.executable).parent.joinpath("wormhole")))
    if wormhole_cmd is not None:
        return wormhole_cmd


@click.command()
@click.option("--system", default="default", help="NETLAB+ system to communicate with.")
@click.option("--output-file", "output_file", type=click.Path())
@click.option("--wormhole", is_flag=True, help="Send logs over magic-wormhole.")
@click.option("--install-wormhole", is_flag=True, help="Install magic-wormhole if needed.")
@click.pass_context
def debug(ctx, system, output_file, wormhole, install_wormhole):
    """
    This will collect debugging data that will be useful for us to assist you.

    It will:

    \b
    - Collect system data
    - Collect config and where it came from
    - Run openssl s_client
    - Connect to a netlab and collects debug traffic
    - Collect `system_status_get`
    - Collect full traceback of all errors
    - Optional: output to file
    - Optional: install magic wormhole
    - Optional: send the results over magic wormhole

    # Using magic wormhole

    https://magic-wormhole.readthedocs.io/en/latest/

    If you cannot easily collect the debug output, you can use the --wormhole
    flag to easily get the debug output at a different computer. Simply install
    magic-wormhole on the other computer and follow the instructions at the end
    if the debug output.

    Example::

    \b
        $ netlab debug --wormhole --install-wormhole
        2024-09-11 21:12:55 [info     ] writing file     path=/tmp/netlab_debug_0y5xltrx.txt
        ─────────────────────────────── Collecting system data ───────────────────────────────
        ───────────────────────────── Checking netlab sdk config ─────────────────────────────
        ───────────────────────────── Running `openssl s_client` ─────────────────────────────
        ──────────────────────────── Finished `openssl s_client` ─────────────────────────────
        ─────────────────────────── Debug data collection complete ───────────────────────────
        Sending 9.7 kB file named 'netlab_debug_0y5xltrx.txt'
        Wormhole code is: 88-headwaters-blowtorch
        On the other computer, please run:

        wormhole receive 88-headwaters-blowtorch

    In this example, you would run `wormhole receive 88-headwaters-blowtorch`
    on the reciving computer.
    """

    logging.config.dictConfig(LOGGING)

    wormhole_cmd = find_wormhole()

    if wormhole_cmd is None and install_wormhole:
        log.info("trying to install magic-wormhole")
        asyncio.run(run_install_wormhole())

        wormhole_cmd = find_wormhole()

    if wormhole and wormhole_cmd is None:
        log.error("wormhole command not found")
        exit(1)

    if output_file is None or wormhole:
        fd, output_file = tempfile.mkstemp(".txt", "netlab_debug_")

    if output_file or wormhole:
        log.info(f"writing file {output_file=}")

        log_config: Any = copy.deepcopy(LOGGING)

        log_config["root"]["handlers"].append("output_file")
        log_config["handlers"]["output_file"] = {
            "class": "logging.FileHandler",
            "formatter": "default",
            "filename": output_file,
            "mode": "w",
        }

        logging.config.dictConfig(log_config)

    asyncio.run(debugger.main(system=system))

    if wormhole:
        asyncio.run(run_wormhole(wormhole_cmd, output_file))



"""
Datatypes for working with a vm.
"""

from typing import Dict, List, Optional, TypedDict


class VMAllocTotal(TypedDict):
    vm_n_vms: int
    "Total number of virtual machines in this pod. Remote PCs of type ABSENT are not counted."
    vm_n_cpu: int
    "Total number of virtual CPUs consumed by this pod."
    vm_mem_mb: int
    "Total amount of RAM allocated to all VMs in this pod (megabytes)."


class VMAllocHostInfo(TypedDict):
    vm_n_vms: int
    "Total number of virtual machines of the pod assigned to this host."
    vm_n_cpu: int
    "Total number of virtual CPUs consumed by this pod on this host."
    vm_mem_mb: int
    "Total amount of RAM (megabytes) allocated by VMs in this pod on this host."
    vh_pra_enabled: bool
    "Proactive Resource Awareness is enabled on this host."
    vh_pra_max_vm: Optional[bool]
    """Maximum number of virtual machines allowed to be scheduled on this host when PRA enabled.
    `None` if PRA is disabled."""
    vh_pra_max_cpu: Optional[bool]
    """Maximum number of virtual CPUs allowed to be scheduled on this host when PRA enabled.
    None if PRA is disabled."""
    vh_pra_max_mem_mb: Optional[bool]
    """Maximum amount of memory that can be scheduled on this host when PRA enabled.
    None if PRA is disabled."""


class VMAlloc(TypedDict):
    """
    VMAlloc contains information about potential VM resource consumption in a given pod.
    """

    total: VMAllocTotal
    """
    Contains total number of virtual machines, virtual CPUs, and memory allocation of the
    entire pod.
    """
    vh_ids: List[int]
    """
    List of vh_id host identifiers of the virtual machine hosts assigned to the pod's VMs.
    May be empty list if the pod has not yet been assigned to any hosts. If list contains
    one host identifier, then all VMs in the pod are assigned to the same host.
    """
    vh_alloc: Dict[str, VMAllocHostInfo]
    """
    Dict containing number of virtual machines, virtual CPUs, and memory allocation for each host
    assigned to the pod. The keys of vh_alloc are vh_id host identifiers; each key is also be a
    member of the vh_ids list. The values of this dict are type VM_Alloc_HostInfo (see below).
    May be an empty dict if the pod has not yet been assigned to any hosts.
    """



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, TypeAlias

from . import compiler
from .types import ExistingPod


@dataclass
class PodOpBase:
    def op(self) -> tuple[compiler.CompiledPod | None, ExistingPod | None]:
        return None, None


@dataclass
class Superceded(PodOpBase):
    pod: compiler.CompiledDeletePod
    by: compiler.CompiledPod

    def __str__(self):
        return f"[Superceded] {self.by.pod_id} ({self.by.get_name()})"


@dataclass
class Absent(PodOpBase):
    pod: compiler.CompiledDeletePod

    def __str__(self):
        return f"[Absent]     {self.pod.pod_id}"


@dataclass
class Delete(PodOpBase):
    pod: compiler.CompiledDeletePod
    existing: ExistingPod

    def op(self) -> tuple[None, ExistingPod]:
        return None, self.existing

    def __str__(self):
        return f"[Delete]     {self.pod.pod_id} ({self.existing.current_name})"


@dataclass
class Rebuild(PodOpBase):
    pod: compiler.CompiledPod
    existing: ExistingPod
    conflicts: list[tuple[str, Any, Any]]

    def op(self) -> tuple[compiler.CompiledPod, ExistingPod]:
        return self.pod, self.existing

    def __str__(self):
        lines = [f"[Rebuild]    {self.pod.pod_id} ({self.pod.get_name()})"]
        for kind, after, before in self.conflicts:
            lines.append(f"  {kind} = {before} -> {after}")
        return "\n".join(lines)


@dataclass
class Present(PodOpBase):
    pod: compiler.CompiledPod
    existing: ExistingPod

    def op(self) -> tuple[None, None]:
        return None, None

    def __str__(self):
        return f"[Present]    {self.pod.pod_id} ({self.pod.get_name()})"


@dataclass
class Create(PodOpBase):
    pod: compiler.CompiledPod

    def op(self) -> tuple[compiler.CompiledPod, None]:
        return self.pod, None

    def __str__(self):
        return f"[Create] {self.pod.pod_id} ({self.pod.get_name()})"


PodOp: TypeAlias = Superceded | Absent | Delete | Rebuild | Present | Create



import re
from typing import Any, Dict, List, Literal, Optional, Union
from uuid import UUID

from .. import enums
from ..datatypes import HDRResult, PCCloneSpec
from ..utils import hdr_result, omit_nones, version_gte
from ._client_protocol import ClientProtocol

bad_vm_name_chars = re.compile("[^a-zA-Z0-9-.]")


class PodApiMixin(ClientProtocol):
    pod_default_props = ["pod_id", "pod_name", "pt_id", "pod_current_state"]
    pod_props = [
        "def_im_name",
        "def_topology_image",
        "def_vlan_map",
        "device_count",
        "devices",
        "pod_acl_enabled",
        "pod_admin_state",
        "pod_adv_settings",
        "pod_auto_net_enabled",
        "pod_auto_net_host_setup",
        "pod_auto_net_host_teardown",
        "pod_cat",
        "pod_csw_base_vlan",
        "pod_csw_id",
        "pod_current_state",
        "pod_desc",
        "pod_id",
        "pod_id",
        "pod_managed",
        "pod_name",
        "pod_origin",
        "pod_res_id",
        "pod_task_id",
        "pod_task_info",
        "pod_task_type",
        "pod_uuid",
        "pt_actions",
        "pt_adv_settings",
        "pt_apc_port_count",
        "pt_apdid",
        "pt_as_port_count",
        "pt_author",
        "pt_build",
        "pt_copyright",
        "pt_csw_port_count",
        "pt_desc",
        "pt_gpdid",
        "pt_id",
        "pt_name",
        "pt_notes",
        "pt_org",
        "pt_origin",
        "pt_pod_max",
        "pt_removable",
        "pt_tabs",
        "pt_url",
        "pt_vlan_pool",
        "remote_pc",
        "remote_pc_count",
        "reservations",
        "sched_image",
        "vm_alloc",
    ]

    pod_types_default_props = ["pt_id", "pt_name", "pt_build"]
    pod_types_props = [
        "pt_id",
        "pt_name",
        "pt_build",
        "pt_gpdid",
        "pt_apdid",
        "pt_desc",
        "pt_index",
        "pt_type",
        "sched_image",
        "def_topology_image",
        "def_im_name",
        "def_vlan_map",
        "pt_tabs",
        "pt_actions",
        "pt_origin",
        "pt_vlan_pool",
        "pt_pod_max",
        "pt_notes",
        "pt_csw_port_count",
        "pt_as_port_count",
        "pt_apc_port_count",
        "pt_build",
        "pt_author",
        "pt_org",
        "pt_copyright",
        "pt_url",
        "pt_removable",
        "pt_ae_pod",
        "pt_adv_settings",
        "remote_pc_count",
        "device_count",
    ]

    def __pc_clone_specs(self, *, user_pc_clone_specs, remote_pc_list, clone_pod_name):
        # Ensure that user_pc_clone_specs is a list of dicts
        if isinstance(user_pc_clone_specs, dict):
            user_pc_clone_specs = [user_pc_clone_specs] * len(remote_pc_list)
        elif user_pc_clone_specs is None:
            user_pc_clone_specs = [{}] * len(remote_pc_list)

        if len(user_pc_clone_specs) != len(remote_pc_list):
            raise ValueError("`user_pc_clone_specs` and `remote_pc_list` must be the same length.")

        # Ensure that the remote_pc_list is in order
        remote_pc_list = sorted(remote_pc_list, key=lambda e: e["pl_index"])

        # Result of this function
        new_vm_out = []

        for pl_index, vms in enumerate(zip(user_pc_clone_specs, remote_pc_list), 1):
            user_vm, existing_vm = vms
            bad_clone_specs = set(user_vm.keys()) - PCCloneSpec.__annotations__.keys()
            if bad_clone_specs:
                raise ValueError("These keys are not allowed: {}".format(bad_clone_specs))

            if user_vm.get("pc_type", existing_vm["pc_type"]) == enums.PCType.ABSENT:
                new_vm_out.append(
                    {
                        "pl_index": pl_index,
                        "pc_type": user_vm.get("pc_type", existing_vm["pc_type"]),
                    }
                )
            else:
                vm_path_match = re.match(r"^\[(.*)\].*", existing_vm["vm_path"])
                assert vm_path_match

                copied_user_vm = user_vm.copy()
                del user_vm

                clone_role = copied_user_vm.pop("clone_role", "NORMAL")
                if clone_role == "TEMPLATE":
                    clone_snapshot = copied_user_vm.pop("clone_snapshot", None)
                else:
                    clone_snapshot = copied_user_vm.pop("clone_snapshot", "GOLDEN")

                clean_clone_name = bad_vm_name_chars.sub("-", clone_pod_name + "-" + existing_vm["pc_label"])
                new_vm_out.append(
                    {
                        "pl_index": pl_index,
                        "pc_type": copied_user_vm.pop("pc_type", existing_vm["pc_type"]),
                        "source_vm_id": existing_vm["vm_id"],
                        "source_snapshot": copied_user_vm.pop("source_snapshot", existing_vm["vm_snapshot"]),
                        "clone_snapshot": clone_snapshot,
                        "clone_type": copied_user_vm.pop("clone_type", "LINKED"),
                        "clone_role": clone_role,
                        "clone_name": copied_user_vm.pop("clone_name", clean_clone_name),
                        "clone_storage_alloc": copied_user_vm.pop("clone_storage_alloc", "ONDEMAND"),
                        "clone_datastore": copied_user_vm.pop("clone_datastore", vm_path_match.group(1)),
                        "clone_vh_id": copied_user_vm.pop("clone_vh_id", existing_vm["vh_id"]),
                        **copied_user_vm,
                    }
                )

        return new_vm_out

    async def pod_update(
        self,
        *,
        pod_id: int,
        pod_name: Optional[str] = None,
        pod_desc: Optional[str] = None,
        pod_auto_net_enabled: Optional[bool] = None,
        pod_auto_net_host_setup: Optional[bool] = None,
        pod_auto_net_host_teardown: Optional[bool] = None,
        pod_adv_settings: Optional[str] = None,
        pod_acl_enabled: Optional[bool] = None,
        **kwargs,
    ) -> None:
        """
        This method allows you to update a pod on the NETLAB+ system.

        :param int pod_id: Unique pod identifier.
        :param str pod_name: Pod name.
        :param str pod_desc: Pod description.
        :param bool pod_auto_net_enabled: Auto-networking enabled.
        :param bool pod_auto_net_host_setup: Auto-networking host setup enabled.
        :param bool pod_auto_net_host_teardown: Auto-networking host teardown enabled.
        :param str pod_adv_settings: Advanced settings for pod configuration.
        :param bool pod_acl_enabled: Pod access control list enabled.
        """

        method = "pod.update"
        param: Dict[str, Any] = {
            "pod_id": pod_id,
            "pod_name": pod_name,
            "pod_desc": pod_desc,
            "pod_auto_net_enabled": pod_auto_net_enabled,
            "pod_auto_net_host_setup": pod_auto_net_host_setup,
            "pod_auto_net_host_teardown": pod_auto_net_host_teardown,
            "pod_adv_settings": pod_adv_settings,
            "pod_acl_enabled": pod_acl_enabled,
        }

        param = omit_nones(
            param,
            {
                "pod_name",
                "pod_desc",
                "pod_auto_net_enabled",
                "pod_auto_net_host_setup",
                "pod_auto_net_host_teardown",
                "pod_adv_settings",
                "pod_acl_enabled",
            },
        )

        return await self.call(method, **param, **kwargs)

    async def pod_add(
        self,
        *,
        pod_id: int,
        pt_id: str,
        devices: Optional[List[Dict[str, Any]]] = None,  # TODO
        pod_acl_enabled: bool = False,
        pod_adv_settings: Optional[str] = None,
        pod_auto_net_enabled: bool = True,
        pod_auto_net_host_setup: bool = True,
        pod_auto_net_host_teardown: bool = True,
        pod_cat: enums.PodCategory = enums.PodCategory.PERSISTENT_VM,
        pod_desc: Optional[str] = None,
        pod_name: Optional[str] = None,
        **kwargs,
    ) -> Literal["OK"]:
        """
        This method allows you to add a pod to the NETLAB+ system. Pods created will be OFFLINE after added.

        :param pod_id: Unique pod identifier.
        :param pt_id: Pod type identifier.
        :param devices: Array of devices associated with this pod. This is for real equipment pods only.
        :param pod_acl_enabled: Pod access control list enabled.
        :param pod_adv_settings: Advanced settings for pod configuration.
        :param pod_auto_net_enabled: Auto-networking enabled.
        :param pod_auto_net_host_setup: Auto-networking host setup enabled.
        :param pod_auto_net_host_teardown: Auto-networking host teardown enabled.
        :param pod_cat: Pod category type.
        :param pod_desc: Pod description.
        :param pod_name: Pod name.
        """

        method = "pod.add"
        param = {
            "pod_id": pod_id,
            "pt_id": pt_id,
            "devices": devices,
            "pod_acl_enabled": pod_acl_enabled,
            "pod_adv_settings": pod_adv_settings,
            "pod_auto_net_enabled": pod_auto_net_enabled,
            "pod_auto_net_host_setup": pod_auto_net_host_setup,
            "pod_auto_net_host_teardown": pod_auto_net_host_teardown,
            "pod_cat": pod_cat,
            "pod_desc": pod_desc,
            "pod_name": pod_name,
        }

        param = omit_nones(
            param,
            {
                "devices",
                "pod_acl_enabled",
                "pod_adv_settings",
                "pod_desc",
                "pod_name",
            },
        )

        param.update(kwargs)
        return await self.call(method, **param)

    async def pod_get(
        self, *, pod_id: int, properties: Union[List[str], Literal["all", "default"]] = "default", **kwargs
    ) -> Dict[str, Any]:
        """

        :param pod_id: Unique pod identifier.
        :param properties: Properties consist of any combination of properties listed below under Properties.
                           You may also use **"all"** to include all properties, or **"default"** for a smaller subset
                           of properties.

        :return: Returns a dictionary object containing the following properties:

        **Properties**

        def_im_name
            Default image name.
        def_topology_image
            Default topology image URL.
        def_vlan_map
            Default VLAN mapping.
        device_count
            Count of devices in pod type.
        devices
            List of devices in pod type.
        pod_acl_enabled
            Pod access control list enabled.
        pod_admin_state
            Administrative pod states.
        pod_adv_settings
            A list of all advanced settings.
        pod_auto_net_enabled
            Automatic VM networking enabled.
        pod_auto_net_host_setup
            Automatic Host networking setup enabled.
        pod_auto_net_host_teardown
            Automatic Host networking teardown enabled.
        pod_cat
            Pod category type. :py:class:`netlab.enums.PodCategory`
        pod_csw_base_vlan
            Pod control switch base VLAN.
        pod_csw_id
            Pod control switch id.
        pod_current_state
            Pod current state status.
        pod_desc
            Pod description.
        pod_id
            Unique pod identifier.
        pod_managed
            Pod is managed.
        pod_name
            Pod name.
        pod_origin
            Pod origin.
        pod_res_id
            Current pod reservation identifier.
        pod_task_id
            Current pod task identifier.
        pod_task_info
            Current pod task information.
        pod_task_type
            Current pod task type.
        pod_uuid
            Globally unique pod identifier.
        pt_actions
            Pod type actions enabled.
        pt_adv_settings
            Long string of advanced settings.
        pt_apc_port_count
            Switched outlet port count requirement.
        pt_apdid
            Author identifier for pod design.
        pt_as_port_count
            Access server port count requirement.
        pt_author
            Pod design author.
        pt_build
            Pod build number.
        pt_copyright
            Pod design copyright.
        pt_csw_port_count
            Control switch port count requirement.
        pt_desc
            Pod type description.
        pt_gpdid
            Global unique identifier.
        pt_id
            Pod type identifier.
        pt_name
            Pod type name.
        pt_notes
            Pod type notes.
        pt_org
            Pod design organization.
        pt_origin
            Pod type origin.
        pt_pod_max
            Max pod count of this type.
        pt_removable
            Pod type is removable.
        pt_tabs
            Pod type tabs enabled.
        pt_url
            Pod design support URL.
        pt_vlan_pool
            Pod type vlan pool.
        remote_pc
            List of remote PCs in pod type with their expanded properties. See Remote PC table below.
        remote_pc_count
            Count of remote PCs in pod type.
        reservations
            The number of reservations scheduled or in progress for this pod.
        sched_image
            Scheduler image URL.
        vm_alloc
            VM resource allocation properties. `netlab.datatypes.VMAlloc` Added in 18.5.0.
        """

        method = "pod.get"

        if properties == "all":
            properties = self.pod_props
        if properties == "default":
            properties = self.pod_default_props

        res_param = {
            "pod_id": pod_id,
            "properties": properties,
        }
        res_param.update(kwargs)
        return await self.call(method, **res_param)

    async def pod_list(self, *, pod_cat: Optional[enums.PodCategory] = None, **kwargs) -> List[Dict[str, Any]]:
        """
        List pods on the system.

        :param pod_cat: Pod category type.
        :return: Returns a list of dictionary objects containing the following properties:

        **Properties**

        def_topology_image
            Default topology image URL.
        pod_acl_enabled
            Pod access control lists enabled for this pod.
        pod_admin_state
            Pod administrative state. :py:class:`netlab.enums.PodAdminState`
        pod_cat
            Pod category. :py:class:`netlab.enums.PodCategory`
        pod_current_state
            Pod current state. :py:class:`netlab.enums.PodCurrentState`
        pod_desc
            Pod description.
        pod_dyn_vlan
            Pod is using dynamic vlans.
        pod_id
            Unique pod identifier.
        pod_managed
            Pod is managed (currently always True).
        pod_name
            Pod name.
        pod_uuid
            Pod name.
        pod_res_id
            Current pod reservation identifier.
        pt_apdid
            Author identifier for pod design.
        pt_desc
            Pod type description.
        pt_gpdid
            Global unique identifier.
        pt_id
            Pod type identifier.
        pt_name
            Pod type name.
        sched_image
            Scheduler image URL.
        """
        method = "pod.list"

        res_param = {}
        if pod_cat is not None:
            res_param["pod_cat"] = pod_cat

        res_param.update(kwargs)
        return await self.call(method, **res_param)

    async def pod_list_used_ids(self, **kwargs) -> List[int]:
        """
        List the used pod_id's on the system.

        :return: Returns a list of pod_ids
        """

        method = "pod.list.used_ids"

        pod_ids = await self.call(method, **kwargs)

        return list(map(int, pod_ids))

    async def pod_pc_get(
        self, *, pod_id: Optional[int] = None, pl_index: Optional[int] = None, pc_id: Optional[int] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get information regarding a pod pc.

        :param pod_id: Pod ID
        :param pl_index: PC Pod index
        :param pc_id: unique pc identifier

        :return: On success, returns the following properties:

        **Properties**

        pc_icon
            Icon identifier. :py:class:`netlab.enums.PCIcon`
        pc_id
            Unique pc identifier.
        pc_label
            PC label.
        pc_online
            PC online.
        pc_origin
            PC origin.
        pc_os_id
            PC operating system identifier.
        pc_os_name
            NETLAB+ operating system name.
        pc_pod_id
            pod_id pc is associated to.
        pc_pod_index
            Index of pods.
        pc_revert_to_scrub
            Revert pc during scrub.
        pc_type
            PC type. :py:class:`netlab.enums.PCType`
        pc_vnc_ip_addr
            IP Address of vnc (ESXi host) server.
        pc_vnc_port
            Port number VNC is running on.
        pc_vnc_use_copy_rect
            Always Allow VNC to use copy rect encoding.
        pl_icon
            Pod design PC icon identifier. :py:class:`netlab.enums.PLIcon`
        pl_index
            PC pod index.
        pl_label
            Remote pc label.
        pt_id
            Pod type identifier.
        vdc_id
            Agent hostname/IP address.
        vdc_name
            VM Datacenter name.
        vh_bios_date
            Timestamp of BIOS creation date.
        vh_bios_version
            BIOS version installed.
        vh_com_path
            Virtual host communication path. :py:class:`netlab.enums.VirtualHostComPath`
        vh_cpu_cores
            Total count of CPU cores on virtual host.
        vh_cpu_mhz
            Speed in MHz of CPU.
        vh_cpu_model
            CPU model identification.
        vh_cpu_threads
            Number of threads CPUs can handle.
        vh_cpu_threads
            Number of threads CPUs can handle.
        vh_date_modified
            Timestamp of virtual host information has been modified.
        vh_date_tested
            Timestamp of virtual host test communication.
        vh_id
            Virtual host identifier.
        vh_inside_ipv4_addr
            Inside interface ipv4 address.
        vh_inside_vswitch_0
            Virtual host inside vSwitch.
        vh_last_test_status
            Status of last virtual host test communication.
        vh_memory_mb
            Total RAM in MB of virtual host.
        vh_name
            Virtual host hostname.
        vh_online
            Virtual host is online.
        vh_os_build
            Operating system build number.
        vh_os_description
            Operating system description.
        vh_os_name
            Operating system name.
        vh_os_type
            Operating system type.
        vh_os_vendor
            Operating system vendor.
        vh_os_version
            Operating system version.
        vh_outside_ipv4_addr
            Outside interface ipv4 address.
        vh_pra_enabled
            Proactive resource awareness is enabled.
        vh_pra_max_cpu
            Max number of CPUs for PRA.
        vh_pra_max_mem_mb
            Max amount of RAM in MB for PRA.
        vh_pra_max_vm
            Max amount of virtual machines for PRA.
        vh_sys_model
            System model.
        vh_sys_service_tag
            System service tag.
        vh_sys_vendor
            System vendor.
        vh_uuid
            Virtual host unique global identifier.
        vhg_id
            Virtual host group identifier.
        vm_alloc_cpu_n
            Total amount of allocated CPUs.
        vm_alloc_mem_mb
            Total amount of allocated RAM.
        vm_auto_display
            Virtual machine auto display port configuration.
        vm_auto_network
            Virtual machine auto networking.
        vm_auto_settings
            Virtual machine auto settings.
        vm_child_count
            Total count of children VMs.
        vm_comments
            Virtual machine comments from VMX.
        vm_date_added
            Timestamp of virtual machine added to NETLAB+ inventory.
        vm_id
            Virtual machine identifier.
        vm_name
            Virtual machine name.
        vm_netlab_os_id
            NETLAB+ match to vm_vendor_os_id.
        vm_parent_id
            Virtual machine parent identifier, if VM is a link clone.
        vm_parent_snapname
            Virtual machine parent snapshot name, if VM is a link clone.
        vm_path
            Virtual machine path on datastore.
        vm_power_state
            Virtual machine power state.
        vm_role
            Virtual machine role in NETLAB+ inventory. :py:class:`netlab.enums.VirtualMachineRole`
        vm_runtime_vh_id
            Unique identifier of runtime virtual host.
        vm_sanity_checks
            Sanity checks.
        vm_shutdown_pref
            Shutdown preference.
        vm_snapshot
            This is what the VM reverts to when the pod is reset. You can take multiple snapshots with the same name,
            without having to update this field. Update this field with a call to
            :meth:`~netlab.api.PodApiMixin.pod_pc_update`
        vm_uuid
            Virtual machine unique identifier.
        vm_vendor_os_id
            Vendor operating system identifier.
        vm_vendor_os_name
            Vendor operating system name.
        """

        method = "pod.pc.get"

        res_param = {}
        if pod_id is not None:
            res_param["pod_id"] = pod_id
        if pl_index is not None:
            res_param["pl_index"] = pl_index
        if pc_id is not None:
            res_param["pc_id"] = pc_id

        res_param.update(kwargs)
        return await self.call(method, **res_param)

    async def pod_pc_update(
        self,
        *,
        pc_id: int,
        pc_type: Optional[Literal["ABSENT", "AVMI"]] = None,
        pc_os_id: Optional[str] = None,
        pc_online: Optional[bool] = None,
        vm_shutdown_pref: Optional[str] = None,
        vm_id: Optional[int] = None,
        vm_snapshot: Optional[str] = None,
        vm_auto_display: Optional[bool] = None,
        vm_auto_network: Optional[bool] = None,
        vm_auto_settings: Optional[bool] = None,
        vm_sanity_checks: Optional[bool] = None,
        **kwargs,
    ) -> None:
        """
        Update information regarding a pod pc.

        :param pc_id: unique pc identifier
        :param pc_type: PC type (ABSENT, AVMI)
        :param pc_os_id: PC operating system identifier
        :param pc_online: PC online
        :param vm_shutdown_pref: Shutdown preference
        :param vm_id: Virtual machine identifier
        :param vm_snapshot: This is what the VM reverts to when the pod is reset. You can take multiple snapshots
                                with the same name, without having to update this field.
        :param vm_auto_display: Virtual machine auto display port configuration
        :param vm_auto_network: Virtual machine auto networking
        :param vm_auto_settings: Virtual machine auto settings
        :param vm_sanity_checks: Virtual machine sanity checks
        """

        method = "pod.pc.update"
        params: Dict[str, Any] = {
            "pc_id": pc_id,
        }

        if pc_type:
            params["pc_type"] = pc_type
        if pc_os_id:
            params["pc_os_id"] = pc_os_id
        if pc_online:
            params["pc_online"] = pc_online
        if vm_shutdown_pref:
            params["vm_shutdown_pref"] = vm_shutdown_pref
        if vm_id:
            params["vm_id"] = vm_id
        if vm_snapshot:
            params["vm_snapshot"] = vm_snapshot
        if vm_auto_display:
            params["vm_auto_display"] = vm_auto_display
        if vm_auto_network:
            params["vm_auto_network"] = vm_auto_network
        if vm_auto_settings:
            params["vm_auto_settings"] = vm_auto_settings
        if vm_sanity_checks:
            params["vm_sanity_checks"] = vm_sanity_checks

        params.update(kwargs)
        return await self.call(method, **params)

    async def pod_state_change(self, *, pod_id: int, state: enums.PodState, **kwargs) -> Literal["OK"]:
        """
        Sets pod state to either ONLINE, OFFLINE or RESUME.

        :param pod_id: Pod ID
        :param state: PC Pod State
        """

        method = "pod.state.change"

        res_param = {
            "pod_id": pod_id,
            "state": state,
        }

        res_param.update(kwargs)
        return await self.call(method, **res_param)

    async def pod_types_get(
        self,
        *,
        pt_id: str,
        properties: Union[List[str], Literal["default"], Literal["all"]] = "default",
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Returns the build, id and name of installed pods on queried system.

        :param pt_id: Pod Type ID is a alpha/num hash with prefix.
        :param properties: Properties consist of any combination of properties listed below under Properties.
                           You may also use **"all"** to include all properties, or **"default"** for a smaller subset
                           of properties.

        :return: Returns all or selected properties from the list initially passed for return.

        **Properties**

        pt_id
            Pod type identifier.
        pt_name
            Account email.
        pt_build
            Account Family(Last) name.
        pt_gpdid
            Global unique identifer.
        pt_apdid
            Author identifier for pod design.
        pt_desc
            Pod type description.
        pt_index
            Pod type index.
        pt_type
            Pod type id
        sched_image
            Location of scheduler image.
        def_topology_image
            Default topology image URL.
        def_im_name
            Default image map name.
        def_vlan_map
            Default VLAN map.
        pt_tabs
            Pod type tabs.
        pt_actions
            Pod type actions.
        pt_origin
            Pod type origin.
        pt_vlan_pool
            Pod type vlan pool.
        pt_pod_max
            Max amount of pod types allowed on system.
        pt_notes
            Pod type notes.
        pt_csw_port_count
            Number of consecutive ports on control switch required by pod type.
        pt_as_port_count
            Number of access server ports required by pod type.
        pt_apc_port_count
            Number of switched outlet ports required by pod type.
        pt_build
            Pod type build number.
        pt_author
            Pod type author.
        pt_org
            Pod type organization.
        pt_copyright
            Pod type copyright.
        pt_url
            Pod type URL address.
        pt_removable
            Allows pod type to be uninstalled.
        pt_ae_pod
            Pod type is for NETLAB+ AE.
        pt_adv_settings
            Pod type advanced settings.
        hotspots
            Returns array of hotspot information for **pt_id**.

            - im_coords
            - im_dev_class
            - im_dev_name
            - im_group
            - im_name
            - im_shape

        remote_pc_layout
            Returns array of remote PC layout information for **pt_id**.

            - pl_access_types
            - pl_def_access_message
            - pl_def_offline_message
            - pl_icon
            - pl_index
            - pl_label
            - pl_os_types
            - pl_pt_id
            - pl_reboot_to_scrub
            - pl_required

        remote_pc_count
            Get a count of the remote PCs per pod type.
        device_layout
            Returns array of device layout information for **pt_id**.

            - dl_pt_id
            - dl_index
            - dl_name
            - dl_class
            - dl_auto
            - dl_managed
            - dl_so_base_ix
            - dl_as_base_ix
            - dl_cs_base_ix
            - dl_eth_recv_port
            - dl_eth_req
            - dl_ser_req
            - dl_ae_ifx

        device_count
            Get count of the devices for pod type.
        pod_cat_values
            Returns array of pod categories (``PodCategory``).
        cables
            Returns list of cable connections for pod type.

            - ca_pt_id
            - ca_from_type
            - ca_from_index
            - ca_from_port
            - ca_seq
            - ca_cable
            - ca_to_dev
            - ca_to_port
        """

        method = "pod.types.get"

        if properties == "all":
            properties = self.pod_types_props
        if properties == "default":
            properties = self.pod_types_default_props

        res_param = {"pt_id": pt_id, "properties": properties}

        res_param.update(kwargs)
        return await self.call(method, **res_param)

    async def pod_types_list(
        self, *, properties: Union[List[str], Literal["all", "default"]] = "default", **kwargs
    ) -> List[Dict[str, Any]]:
        """
        Obtain a list of pod types.

        :param properties: Properties consist of any combination of properties listed below under Properties.
                           You may also use **"all"** to include all properties, or **"default"** for a smaller subset
                           of properties.
        :return: On success, returns [rt_id, pt_build, pt_name] by default, or the following properties if requested.

        **Properties**

        pt_id
            Pod type identifier.
        pt_name
            Account email.
        pt_build
            Account Family(Last) name.
        pt_gpdid
            Global unique identifer.
        pt_apdid
            Short unique identifier.???
        pt_desc
            Pod type description.
        pt_index
            Pod type index.
        pt_type
            ???
        sched_image
            Location of scheduler image.
        def_topology_image
            Location of the default topology image.
        def_im_name
            Default image map name.
        def_vlan_map
            Default VLAN map.
        pt_tabs
            Pod type tabs.
        pt_actions
            Pod type actions.
        pt_origin
            Pod type origin.
        pt_vlan_pool
            Pod type vlan pool.
        pt_pod_max
            Max amount of pod types allowed on system.
        pt_notes
            Pod type notes.
        pt_csw_port_count
            Number of consecutive ports on control switch required by pod type.
        pt_as_port_count
            Number of access server ports required by pod type.
        pt_apc_port_count
            Number of switched outlet ports required by pod type.
        pt_build
            Pod type build number.
        pt_author
            Pod type author.
        pt_org
            Pod type organization.
        pt_copyright
            Pod type copyright.
        pt_url
            Pod type URL address.
        pt_removable
            Allows pod type to be uninstalled.
        pt_ae_pod
            Is pod type for NETLAB+ AE???.
        pt_adv_settings
            Pod type advanced settings.
        remote_pc_count
            Get count of the remote PCs per pod type.
        device_count
            Get count of the devices per pod type.

        """
        method = "pod.types.list"

        if properties == "all":
            properties = self.pod_types_props
        if properties == "default":
            properties = self.pod_types_default_props

        return await self.call(method, properties=properties, **kwargs)

    async def pod_remove_task(
        self,
        *,
        pod_id: int,
        remove_vms: enums.RemoveVMS = enums.RemoveVMS.NONE,
        severity_level: enums.HDRSeverity = enums.HDRSeverity.WARN,
        **kwargs,
    ) -> HDRResult:
        """
        :param pod_id: Pod unique identifier.
        :param remove_vms: Specifies the method for removing the virtual machines.
        :param severity_level: Display detailed events for the task only at this severity level or
            higher.

        :return: HDR result
        """

        method = "pod.remove.task"

        params = {
            "pod_id": pod_id,
            "remove_vms": remove_vms,
            "notify_progress": False,
            "notify_complete": False,
        }

        params.update(kwargs)
        result = await self.call(method, **params)

        if version_gte(self._server_version, "17.1.4"):
            return hdr_result(result, severity_level)
        else:
            return result

    async def pod_clone_task(
        self,
        *,
        source_pod_id: int,
        clone_pod_id: int,
        clone_pod_name: str,
        pc_clone_specs: Optional[Union[List[PCCloneSpec], PCCloneSpec]] = None,
        severity_level: enums.HDRSeverity = enums.HDRSeverity.WARN,
        **kwargs,
    ) -> HDRResult:
        """
        Clone a pod.

        :param source_pod_id: Source pod unique identifier.
        :param clone_pod_id: Cloned pod unique identifier
        :param clone_pod_name: Cloned unique pod name.
        :param pc_clone_specs: Array of PC clone specifications. Any unpassed params will be taken from the existing
            pods. If a dict is passed, the params will be used on every pod.
        :param severity_level: Display detailed events for the task only at this severity level or
            higher.
        :return: HDR result

        **pc_clone_specs**

        The format requires a list of dictionaries.
        One dictionary per each PC in the pod, e.g. `[{pc.clone.spec.1}, {pc.clone.spec.2}]`.
        See ``PCCloneSpec`` the expected value type and a key description.

        **Example:**

        ::

            pc_clone_specs = [{
                    "clone_role":"NORMAL",
                    "clone_type":"LINKED",
                    "source_snapshot":"GOLDEN_MASTER",
                    "pc_type":"AVMI",
                    "clone_vh_id":1,
                    "clone_name":"Test_Linux_Install_pod04_CentOS Server",
                    "clone_storage_alloc":"ONDEMAND",
                    "clone_datastore":"local",
                    "clone_snapshot":"GOLDEN_MASTER"
                }, {
                    "clone_role":"NORMAL",
                    "clone_type":"LINKED",
                    "source_snapshot":"GOLDEN_MASTER",
                    "pc_type":"AVMI",
                    "clone_vh_id":1,
                    "clone_name":"Test_Linux_Install_pod04_Ubuntu Workstation",
                    "clone_storage_alloc":"ONDEMAND",
                    "clone_datastore":"local",
                    "clone_snapshot":"GOLDEN_MASTER"
                }]

        """

        method = "pod.clone.task"

        remote_pc = await self.pod_get(pod_id=source_pod_id, properties=["remote_pc"])

        pc_clone_specs = self.__pc_clone_specs(
            clone_pod_name=clone_pod_name, user_pc_clone_specs=pc_clone_specs, remote_pc_list=remote_pc["remote_pc"]
        )

        result = await self.call(
            method,
            source_pod_id=source_pod_id,
            clone_pod_id=clone_pod_id,
            clone_pod_name=clone_pod_name,
            pc_clone_specs=pc_clone_specs,
            **kwargs,
        )

        if version_gte(self._server_version, "17.1.4"):
            return hdr_result(result, severity_level)
        else:
            return result

    async def pod_acl_add(
        self,
        *,
        com_id: int,
        pod_id: int,
        acc_id: Optional[int] = None,
        cls_id: Optional[int] = None,
        team: Optional[str] = None,
        **kwargs,
    ) -> UUID:
        """
        This method allows you to add access controls to a pod on the NETLAB+ system.

        :param com_id: Community identifier.
        :param pod_id: Pod identifier.
        :param acc_id: Account identifier.
        :param cls_id: Class identifier.
        :param team: Team assignment.
        :return: Pod access control list UUID.
        """

        args = {
            "com_id": com_id,
            "pod_id": pod_id,
            "acc_id": acc_id,
            "cls_id": cls_id,
            "team": team,
        }
        args.update(kwargs)
        return await self.call("pod.acl.add", **args)

    async def pod_acl_admin_add(self, *, acc_id: int, pod_id: int, **kwargs) -> None:
        """
        This method allows you add additional pod access control administrators.

        :param int acc_id: Account identifier.
        :param int pod_id: Pod identifier.
        """

        return await self.call("pod.acl.admin.add", acc_id=acc_id, pod_id=pod_id, **kwargs)

    async def pod_acl_admin_auth(
        self,
        *,
        acc_id: int,
        pod_id: int,
        **kwargs,
    ) -> bool:
        """
        This method returns true if a specified instructor account is authorized to manage POD
        ACLs on a specified pod.

        :param acc_id: Account identifier.
        :param pod_id: Pod identifier.

        :return: acc_id is authorized for pod_id.
        """

        result = await self.call("pod.acl.admin.auth", acc_id=acc_id, pod_id=pod_id, **kwargs)

        return bool(int(result))

    async def pod_acl_admin_list(self, *, pod_id: int, **kwargs) -> List[Dict[str, Any]]:
        """
        This method allows you to retrieve a list of pod access controls per pod.

        :param pod_id: Pod identifier.
        :return: Pod access control list.
        """

        return await self.call("pod.acl.admin.list", pod_id=pod_id, **kwargs)

    async def pod_acl_admin_pods(self, *, acc_id: int, **kwargs) -> List[Dict[str, Any]]:
        """
        This method allows you to retrieve a list of pods a user has access to.

        :param int acc_id: Account identifier.
        """

        return await self.call("pod.acl.admin.pods", acc_id=acc_id, **kwargs)

    async def pod_acl_admin_remove(self, *, acc_id: int, pod_id: int, **kwargs) -> None:
        """
        This method allows you to remove authorization for an account.

        :param acc_id: Account identifier.
        :param pod_id: Pod identifier.
        """

        return await self.call("pod.acl.admin.remove", acc_id=acc_id, pod_id=pod_id, **kwargs)

    async def pod_acl_challenge(
        self,
        *,
        com_id: int,
        pod_id: int,
        acc_id: Optional[int] = None,
        cls_id: Optional[int] = None,
        team: Optional[str] = None,
        **kwargs,
    ) -> UUID:
        """
        This method allows you to test against a pod access control list.

        :param int com_id: Community identifier.
        :param int pod_id: Pod identifier.
        :param int acc_id: Account identifier.
        :param int cls_id: Class identifier.
        :param str team: Team assignment.
        :return: If matching rule found, pod access control list UUID.
        """

        return await self.call(
            "pod.acl.challenge", com_id=com_id, pod_id=pod_id, acc_id=acc_id, cls_id=cls_id, team=team, **kwargs
        )

    async def pod_acl_get(self, *, pacl_uuid: UUID, **kwargs) -> List[Dict[str, Any]]:
        """
        This method allows you to retrieve a pod access control list.

        :param pacl_uuid: Pod access control list UUID.
        :return: Pod access control list.
        """

        return await self.call("pod.acl.get", pacl_uuid=pacl_uuid, **kwargs)

    async def pod_acl_list(self, *, pod_id: int, **kwargs) -> List[Dict[str, Any]]:
        """
        This method allows you to retrieve a list of pod access controls per pod.

        :param pod_id: Pod identifier.
        :return: Pod access control list.
        """

        return await self.call("pod.acl.list", pod_id=pod_id, **kwargs)

    async def pod_acl_remove(self, *, pacl_uuid: UUID, **kwargs) -> None:
        """
        This method allows you to remove a pod access control list.

        :param pacl_uuid: Pod access control list UUID.
        """

        return await self.call("pod.acl.remove", pacl_uuid=pacl_uuid, **kwargs)



from datetime import date as date_type
from datetime import datetime, timedelta
from typing import Any, Dict, List, Literal, Optional, overload

from .. import enums
from ..utils import minimum_version, omit_nones
from ._client_protocol import ClientProtocol


class ReservationApiMixin(ClientProtocol):
    @staticmethod
    def reservation_timeslot_adjuster(*, user_time: datetime, round_low: bool = False) -> datetime:
        """
        Round a datetime to the nearest half hour, which is required for some times in netlab.

        :param user_time: Must be a datetime object. Adjusted for fulfilling time slot
            if not already time slot acceptable.
        :param round_low: True: rounds time in minutes down to the low timeslot. False: rounds time up to next
            timeslot.

        :return: Datetime rounded to half hour.
        """

        adjusted_minutes = 30 * (user_time.minute // 30)  # floor division used to round minutes down
        adjusted_time = user_time.replace(minute=adjusted_minutes, second=0, microsecond=0)

        if adjusted_time == user_time or round_low:
            pass
        else:
            adjusted_time += timedelta(minutes=30)

        return adjusted_time

    async def reservation_make(
        self,
        *,
        type: enums.ReservationType,
        pod_id: int,
        cls_id: int,
        end_time: datetime,
        start_time: Optional[datetime] = None,
        acc_id: Optional[int] = None,
        reserver_id: Optional[int] = None,
        tz_id: Optional[int] = None,
        tz_olson: Optional[str] = None,
        team: Optional[str] = None,
        pt_id: Optional[int] = None,
        ex_id: str,
        init_config: enums.InitConfig = enums.InitConfig.NONE,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        :param type: Type of reservation.
        :param pod_id: Pod identifier.
        :param cls_id: Class identifier.
        :param end_time: Local time end of reservation. This must be on a half hour slot boundary.
        :param start_time: Local time start of reservation. If not specified it will use the current
            datetime.
        :param tz_id: Timezone identifier.
        :param tz_olson: IANA timezone name. If not specified, the timezone of the user identified by acc_id will be
            used.
        :param team: Team identifier.
        :param pt_id: Pod type identifier.
        :param ex_id: Exercise identifier **(required)**.
        :param acc_id: Account identifier for the reservation. Defaults to **reserver_id**.
        :param init_config: Initial configuration.
        :param reserver_id: Reserver identifier. Defaults to the request user. An admin can use a **reserver_id** to
            make reservations for others.

        :return: On success, returns an object with the following properties:

        **Properties**

        res_id
            Reservation identifier.
        res_start_utc
            Reservation start in UTC.
        res_end_utc
            Reservation end in UTC.
        res_minutes
            Reservation time in minutes
        """

        method = "reservation.make"

        assert isinstance(type, enums.ReservationType)

        # Because for some reason these are diffrent
        if type is enums.ReservationType.INDIVIDUAL:
            fixed_type = "INDIVIDUAL"
        elif type is enums.ReservationType.ILT_CLASS:
            fixed_type = "ILT_CLASS"
        elif type is enums.ReservationType.INSTRUCTOR:
            fixed_type = "INSTRUCTOR"
        elif type is enums.ReservationType.TEAM:
            fixed_type = "TEAM"

        res_param = {
            "type": fixed_type,
            "pod_id": pod_id,
            "end_time": end_time,
            "start_time": start_time,
            "acc_id": acc_id,
            "reserver_id": reserver_id,
            "tz_id": tz_id,
            "tz_olson": tz_olson,
            "team": team,
            "pt_id": pt_id,
            "ex_id": ex_id,
            "cls_id": cls_id,
            "init_config": init_config,
        }

        res_param = omit_nones(
            res_param,
            {
                "start_time",
                "acc_id",
                "reserver_id",
                "tz_id",
                "tz_olson",
                "team",
                "pt_id",
                "ex_id",
                "cls_id",
                "init_config",
            },
        )

        res_param.update(kwargs)
        return await self.call(method, **res_param)

    async def reservation_plan(
        self,
        *,
        acc_id: Optional[int] = None,
        cls_id: Optional[int] = None,
        date: date_type,
        ex_id: Optional[int] = None,
        pod_id: Optional[int] = None,
        reserver_id: Optional[int] = None,
        team: Optional[int] = None,
        type: enums.ReservationType,
        tz_id: Optional[int] = None,
        tz_olsen: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        Used to provide information on all pods available for a proposed date.

        :param acc_id: Account identifier.
        :param cls_id: Class identifier.
        :param date: Date.
        :param ex_id: Exercise identifier.
        :param pod_id: Pod identifier.
        :param reserver_id: Reserver identifier.
        :param team: Team identifier.
        :param type: Type of reservation.
        :param tz_id: Timezone identifier.
        :return: On success, returns an object with the following properties:

        **Properties**

        clock_pref
            User clock preference.
        show_date
            Proposed date.
        min_date
            cls_id start date or SystemMinDate.
        max_date
            cls_id end date or SystemMaxDate.
        time_now
            Current time.
        time_format
            User preferred time format.
        first_weekday
            User preferred first day of the week.
        serial
            Incremented number when there is a scheduler transaction.
        slot_minutes
            Number of minutes in reservation slot.
        time_slots
            Array of available time slots.
        pods
            Array of pods available.
        """
        method = "reservation.plan"

        params = {
            "date": date,
            "type": type,
            "acc_id": acc_id,
            "cls_id": cls_id,
            "ex_id": ex_id,
            "pod_id": pod_id,
            "reserver_id": reserver_id,
            "team": team,
            "tz_id": tz_id,
            "tz_olson": tz_olsen,
        }

        params = omit_nones(
            params,
            {
                "acc_id",
                "cls_id",
                "ex_id",
                "pod_id",
                "reserver_id",
                "team",
                "tz_id",
                "tz_olson",
            },
        )

        return await self.call(method, **params, **kwargs)

    async def reservation_get(self, *, res_id: int, **kwargs) -> Dict[str, Any]:
        """
        Get parameters of a reservation.

        :param int res_id: Reservation identifier.
        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        acc_com_id
            Community identifier.
        acc_full_name
            Account display name.
        acc_id
            Account identifier.
        cls_div_id
            Community identifier.
        cls_id
            Class identifier.
        cls_name
            Class name.
        ex_id
            Lab exercise identifier.
        ex_name
            Lab exercise name.
        pod_desc
            Pod description.
        pod_id
            Unique pod identifier.
        pod_name
            Pod name.
        pt_desc
            Pod type description.
        pt_id
            Pod type identifier.
        pt_name
            Account email.
        res_acc_id
            Account ID.
        res_active_time
            Active reservation timestamp.
        res_asm_data
            ???
        res_attend_time
            Time reservation was attended.
        res_cls_id
            Class identifier.
        res_com_id
            Community identifier.
        res_config
            Initial configuration.
        res_done
            Reservation is done.
        res_end
            Timestamp scheduled end of reservation.
        res_ex_id
            Reservation lab exercise identifier.
        res_flags
            ???
        res_id
            Reservation identifier.
        res_init_fail
            Count of failures during initialization.
        res_is_active
            Reservation is active.
        res_minutes
            Count of minutes for reservation.
        res_pod_id
            Reservation pod identifier.
        res_post_time
            Timestamp of actual end of reservation.
        res_preload_time
            Timestamp UTC lab will preload.
        res_pt_id
            Reservation pod type identifier.
        res_remaining_dhms
            Days, hours, min, sec remaining if pod is in ACTIVE_LOAD or ACTIVE_INIT state.
        res_remaining_sec
            Seconds remaining if pod is in ACTIVE_LOAD or ACTIVE_LAB state.
        res_start
            Timestamp to start reservation.
        res_team
            Reservation team setting.
        res_type
            Reservation type. :py:class:`netlab.enums.ReservationType`
        res_uuid
            Reservation unique identifier.
        sched_image
            Location of the scheduler image.
        """

        method = "reservation.get"
        res_param = {"res_id": res_id}

        res_param.update(kwargs)
        return await self.call(method, **res_param)

    async def reservation_post(self, *, res_id: int, **kwargs) -> None:
        """
        Ends a reservation that has already started.

        :param res_id: Reservation identifier.
        """

        method = "reservation.post"
        return await self.call(method, res_id=res_id, **kwargs)

    async def reservation_cancel(self, *, res_id: int, **kwargs) -> None:
        """
        Cancel a reservation in NETLAB+ prior to reservation being started.

        NOTE - if you wish to end a reservation that is already started use reservation_post.

        :param res_id: Reservation unique identifier.
        """

        method = "reservation.cancel"
        return await self.call(method, res_id=res_id, **kwargs)

    async def reservation_summary(self, **kwargs) -> Dict[str, Any]:
        """
        This method allows you to retrieve a summary of all reservations.

        :return: On success, returns an object with the following properties:

        **Properties**

        active_reservations
            Count of active reservations.
        completed_reservations
            Count of completed reservations.
        future_reservations
            Count of future reservations.
        pods_in_use
            Count of pods in use.
        total_reservations
            Count of total reservations.

        """

        method = "reservation.summary"
        return await self.call(method, **kwargs)

    @overload
    async def reservation_query(
        self,
        *,
        active: Optional[bool] = None,
        min_time: Optional[datetime] = None,
        max_time: Optional[datetime] = None,
        cls_id: Optional[int] = None,
        pod_id: Optional[int] = None,
    ) -> List[Dict[str, Any]]: ...

    @overload
    async def reservation_query(
        self,
        *,
        active: Optional[bool] = None,
        scope: enums.ReservationScope = enums.ReservationScope.ALL,
        min_time: Optional[datetime] = None,
        max_time: Optional[datetime] = None,
        com_id: Optional[int] = None,
        pod_id: Optional[int] = None,
    ) -> List[Dict[str, Any]]: ...

    async def reservation_query(
        self,
        *,
        active: Optional[bool] = None,
        scope: enums.ReservationScope = enums.ReservationScope.ALL,
        min_time: Optional[datetime] = None,
        max_time: Optional[datetime] = None,
        com_id: Optional[int] = None,
        cls_id: Optional[int] = None,
        pod_id: Optional[int] = None,
        **kwargs,
    ) -> List[Dict[str, Any]]:
        """
        Return a list of reservations filtered on criteria. Take note of the acceptable overrides.

        :param scope: Reservation scope.
        :param active: Filter active reservations.
        :param cls_id: Class identifier.
        :param com_id: Community identifier.
        :param max_time: Maximum time range.
        :param min_time: Minimum time range.
        :param pod_id: Pod identifier.

        :return: On success, returns an object with the following properties:

        **Properties**

        acc_com_id
            Community identifier.
        acc_full_name
            Account display name.
        acc_id
            Account identifier.
        cls_div_id
            Community identifier.
        cls_id
            Class identifier.
        cls_name
            Class name.
        ex_id
            Lab exercise identifier.
        ex_name
            Lab exercise name.
        pod_desc
            Pod description.
        pod_id
            Unique pod identifier.
        pod_name
            Pod name.
        pt_desc
            Pod type description.
        pt_id
            Pod type identifier.
        pt_name
            Account email.
        res_acc_id
            Account identifier.
        res_active_time
            Active reservation timestamp.
        res_attend_time
            Time reservation was attended.
        res_cls_id
            Class identifier.
        res_com_id
            Community identifier.
        res_config
            Initial configuration.
        res_done
            Reservation is done.
        res_end
            Timestamp scheduled end of reservation.
        res_ex_id
            Reservation lab exercise identifier.
        res_flags
            Reservation hex code for pod state.
        res_id
            Reservation identifier.
        res_init_fail
            Count of failures during initialization.
        res_is_active
            Reservation is active.
        res_minutes
            Count of minutes for reservation.
        res_pod_id
            Reservation pod identifier.
        res_post_time
            Timestamp of actual end of reservation.
        res_pt_id
            Reservation pod type identifier.
        res_start
            Timestamp to start reservation.
        res_team
            Reservation team setting.
        res_type
            Type of reservation. :py:class:`netlab.enums.ReservationType`
        res_uuid
            Reservation unique identifier.
        sched_image
            Location of the scheduler image.
        """

        method = "reservation.query"

        assert not (cls_id is not None and com_id is not None)

        if com_id is not None:
            assert cls_id is None
            scope = enums.ReservationScope.CLASS

        res_param: Dict[str, Any] = {}
        if active:
            res_param["active"] = active
        if cls_id:
            res_param["cls_id"] = cls_id
        if com_id:
            res_param["com_id"] = com_id
        if max_time:
            res_param["max_time"] = max_time
        if min_time:
            res_param["min_time"] = min_time
        if pod_id:
            res_param["pod_id"] = pod_id
        if scope:
            res_param["scope"] = scope

        res_param.update(kwargs)
        return await self.call(method, **res_param)

    async def reservation_time_now(
        self, *, tz_id: Optional[int] = None, tz_olson: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        This method returns the current NETLAB time.

        :param tz_id: Timezone identifier.
        :param tz_olson:

        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        local
            Local date time.
        local_am_pm
            AM/PM indicator.
        local_day
            Local day of the month.
        local_day_of_week
            Local day of the week.
        local_hmap
            Local time and am/pm.
        local_hms
            Local time.
        local_hour_12
            Local 12 Hour time.
        local_hour_24
            Local 24 Hour time.
        local_minute
            Local minute time.
        local_second
            Local second time.
        local_year
            Local year.
        local_ymd
            Local YYYY-MM-DD.
        tz_id
            Timezone identifier.
        tz_is_dst
            Timezone is in Daylight Savings Time.
        tz_name
            Name of the timezone.
        tz_olson
            Timezone olson.
        utc
            Coordinated Universal Time.
        """

        method = "reservation.time.now"
        res_param: Dict[str, Any] = {}

        if not (tz_id is None or tz_id == 0):  # tz_id==0 is utc, but this fails if passed to api
            res_param["tz_id"] = tz_id
        if tz_olson is not None:
            res_param["tz_olson"] = tz_olson

        res_param.update(kwargs)
        return await self.call(method, **res_param)

    async def reservation_time_offset(
        self,
        *,
        time: Optional[datetime] = None,
        op: Optional[Literal["+", "-"]] = None,
        tz_id_in: Optional[int] = None,
        tz_olson_in=None,
        tz_id_out=None,
        tz_olson_out=None,
        slot_roundup=False,
        years=None,
        months=None,
        weeks=None,
        days=None,
        hours=None,
        minutes=None,
        seconds=None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        This method returns the current time with a timezone offset.

        :param datetime time: Time input.
        :param str op: Operator. "+" rounds up. "-" rounds down.
        :param int tz_id_in: Timezone identifier input.
        :param str tz_olson_in: Timezone olson input.
        :param int tz_id_out: Timezone identifier output.
        :param str tz_olson_out: Timezone olson output.
        :param bool slot_roundup: Round up to nearest slot. i.e. 30 min interval. Defaults to False.
        :param int years: Number of years to increase/decrease.
        :param int months: Number of months to increase/decrease.
        :param int weeks: Number of weeks to increase/decrease.
        :param int days: Number of days to increase/decrease.
        :param int hours: Number of hours to increase/decrease.
        :param int minutes: Number of minutes to increase/decrease.
        :param int seconds: Number of seconds to increase/decrease.

        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        days
            int
        hours
            int
        minutes
            int
        months
            int
        time_in_local
            datetime
        time_in_utc
            datetime
        time_out_local
            datetime
        time_out_utc
            datetime
        tz_id_in
            int
        tz_id_out
            int
        tz_olson_in
            str
        tz_olson_out
            str
        """

        method = "reservation.time.offset"
        res_param = {"slot_roundup": slot_roundup}

        if time:
            res_param["time"] = time
        if op:
            res_param["op"] = op
        if not (tz_id_in is None or tz_id_in == 0):  # tz_id==0 is utc, but this fails if passed to api
            res_param["tz_id_in"] = tz_id_in
        if tz_olson_in:
            res_param["tz_olson_in"] = tz_olson_in
        if not (tz_id_out is None or tz_id_out == 0):  # tz_id==0 is utc, but this fails if passed to api
            res_param["tz_id_out"] = tz_id_out
        if tz_olson_out:
            res_param["tz_olson_out"] = tz_olson_out
        if years:
            res_param["years"] = years
        if months:
            res_param["months"] = months
        if weeks:
            res_param["weeks"] = weeks
        if days:
            res_param["days"] = days
        if hours:
            res_param["hours"] = hours
        if minutes:
            res_param["minutes"] = minutes
        if seconds:
            res_param["seconds"] = seconds

        res_param.update(kwargs)
        return await self.call(method, **res_param)

    async def reservation_time_delta(
        self,
        *,
        end: datetime,
        start: Optional[datetime] = None,
        tz_id: Optional[int] = None,
        tz_olson: Optional[str] = None,
        **kwargs,
    ) -> Dict[str, Any]:
        """
        This method returns time between two timestamps with optional timezone offset.

        :param datetime.datetime end: End time. This must be on a half hour increment.
        :param datetime.datetime start: Start time. Defaults to now if not specified.
        :param int tz_id: Timezone identifier.
        :param str tz_olson: Timezone olson.

        :return: On success, returns an object with the following properties:
        :rtype: dict

        **Properties**

        days
            Number of days to increase/decrease.
        hours
            Number of hours to increase/decrease.
        minutes
            Number of minutes to increase/decrease.
        months
            Number of months to increase/decrease.
        """
        method = "reservation.time.delta"

        res_param: Dict[str, Any] = {
            "end": end,
        }
        if start:
            res_param["start"] = start
        if not (tz_id is None or tz_id == 0):  # tz_id==0 is utc, but this fails if passed to api
            res_param["tz_id"] = tz_id
        if tz_olson:
            res_param["tz_olson"] = tz_olson

        res_param.update(kwargs)
        return await self.call(method, **res_param)

    async def reservation_extend(self, *, res_id, **kwargs) -> Dict[str, Any]:
        """
        Request a 30 minute lab reservation extension.

        :param int res_id: Reservation identifier.

        :return: Returns an object with the following properties:
        :rtype: dict

        **Properties**

        status
            Status code.
        reason
            Textual message that is suitable for display (English).

        **Status Codes**

        OK
            The reservation was extended.
        CLASS_LIMIT_EXCEEDED
            Reservation cannot be extended because of **cls_ext_slots_per_res** setting limit.
        COMMUNITY_LIMIT_EXCEEDED
            Reservation cannot be extended because of **com_ext_slots_per_res** setting limit.
        TIME_SLOT_UNAVAILABLE
            The next time slot is occupied by another reservation.
        POD_ACTIVE_MAX
            Maximum number of pods in use.
        RESERVATION_ALREADY_COMPLETED
            The reservation has already completed and cannot be extended.
        E_ACCESS_DENIED
            Student/learner cannot extend ILT reservation.
        STATE_NOT_ACTIVE_LAB
            Reservation is not in the active lab state.
        TOO_EARLY
            Cannot extend before T-00:15:59 minutes/seconds.
        TOO_LATE
            Cannot extend after T-00:00:10 seconds.

        **Restrictions For All Users**

        + An extension cannot be made before T-00:15:59 from the reservation post time.  This is about 26 minutes from
          the blocked end time and 16 minutes of the usable time remaining.
        + The next 30 minute time slot must not be already scheduled by another reservation.
        + The extension cannot exceed maximum pods in use limit and/or proactive resource awareness settings.

        **Restrictions for Students and Teams**

        + The number of extensions that a student or team can request are limited by community and class settings.
        + For backward setting compatibility, communities and classes do not allow extensions to be made by students or
          teams by default.

            - The administrator must specifically allow extensions for students and teams per community via the
              **com_ext_slots_per_res** setting.
            - The instructor must specifically allow extensions for students and teams per class via the
              **cls_ext_slots_per_res** setting.
            - The number of extensions that a student or team can request in a single reservation will be the more
              restrictive of the two settings.
        + A student (or instructor learner) may not extend an ILT reservation.  Only a lead instructor can do that.
        """

        method = "reservation.extend"
        return await self.call(method, res_id=res_id, **kwargs)

    @minimum_version("18.1.0")
    async def history_reservation_get(self, *, res_id, **kwargs) -> Dict[str, Any]:
        method = "history.reservation.get"
        return await self.call(method, res_id=res_id, **kwargs)



import warnings

from .sync_client import SyncClient as Client

warnings.filterwarnings("default", category=DeprecationWarning)

warnings.warn(
    "'Client' is deprecated. Use 'netlab.sync_client.SyncClient' or 'netlab.async_client.NetlabClient'",
    DeprecationWarning,
)

__all__ = ["Client"]



from enum import Enum


class FreeAlarm(Enum):
    "These values are triggered when a disk's free space has dropped below a determined threshold."

    CLEAR = "CLEAR"
    "The disk has adequate free space."
    WARN = "WARN"
    "The disk has reached free space warning"
    MAJOR = "MAJOR"
    "The disk is nearly out of free space."
    CRITICAL = "CRITICAL"
    "The disk is out of free space."



"""
Netlab enums.
"""

from .common import DateFormat, HDRSeverity, HDRStatus, TimeFormat
from .klass import ClassEmailLogs, ClassExtensionSlots, ClassLabLimit, ContentAccessiblity
from .pod import (
    PCIcon,
    PCType,
    PLIcon,
    PodAdminState,
    PodCategory,
    PodCurrentState,
    PodState,
    RemoveVMS,
    VirtualHostComPath,
    VirtualMachineRole,
)
from .reservation import InitConfig, ReservationScope, ReservationType
from .system import FreeAlarm
from .user import AccountPrivileges, AccountType, CommunityExtensionSlots
from .vm import CloneStorageAllocation, CloneType, PowerState, VirtualMachineLicenseType

__all__ = (
    "TimeFormat",
    "DateFormat",
    "HDRSeverity",
    "HDRStatus",
    "ClassEmailLogs",
    "ClassLabLimit",
    "ClassExtensionSlots",
    "PCIcon",
    "PCType",
    "PLIcon",
    "PodCategory",
    "PodState",
    "RemoveVMS",
    "VirtualHostComPath",
    "VirtualMachineRole",
    "PodAdminState",
    "PodCurrentState",
    "ReservationType",
    "ReservationScope",
    "InitConfig",
    "FreeAlarm",
    "AccountPrivileges",
    "AccountType",
    "CommunityExtensionSlots",
    "ContentAccessiblity",
    "PowerState",
    "CloneStorageAllocation",
    "CloneType",
    "VirtualMachineLicenseType",
)



import json
import sys

from paepy.ChannelDefinition import CustomSensorResult  # type: ignore

from netlab.client import Client

if __name__ == "__main__":
    # interpret first command line parameter as json object
    if len(sys.argv) == 1:
        # for command line debugging, use the default system
        sys_args = '{"params": "default"}'
    else:
        sys_args = sys.argv[1]
    prtg = json.loads(sys_args)
    system = prtg.get("params", "default").lower()

    # PRTG runs sensors as a different user. Specify the Administrator user's netlab config file here
    client = Client(system, config_path="C:/Users/Administrator/.netlab/config.json")

    # query data from netlab api
    response = client.system_usage_cpu()

    # create sensor result
    result = CustomSensorResult("OK")

    # add primary channel for "all" CPU
    total = response.pop("all")
    total_load = 100 - total["idle_pct"]
    result.add_channel(
        channel_name="Total", primary_channel=True, unit="Percent", value=float(total_load), is_float=True
    )

    # additional channel for each individual CPU
    for cpu, values in response.items():
        cpu_load = 100 - values["idle_pct"]
        result.add_channel(
            channel_name="Processor {}".format(cpu), unit="Percent", value=float(cpu_load), is_float=True
        )

    # print sensor result to stdout
    print(result.get_json_result())



from .common import NetlabError, _add


@_add
class ContentNotFoundError(NetlabError):
    """con_id is not found."""

    str_error = "E_CONTENT_NOT_FOUND"


@_add
class ExerciseNotFoundError(NetlabError):
    """ex_id is not found."""

    str_error = "E_EXERCISE_NOT_FOUND"



"""
A sync shim for compatiblity with older versions of the Netlab SDK.
"""

import asyncio
from typing import Any, Dict, Optional, Union

from .async_client import NetlabClient
from .config import NetlabServerConfig


def _sync_wrap_iter(it, loop):
    try:
        while True:
            r = loop.run_until_complete(it.__anext__())
            yield r
    except StopAsyncIteration:
        pass


class SyncClient(object):
    """
    A client instance with which you can use to interact with a remote NETLAB+ system.
    """

    _async_client: NetlabClient

    def __init__(
        self,
        system: str = "default",
        config: Union[None, NetlabServerConfig, Dict[str, Any]] = None,
        config_path: Optional[str] = None,
    ):
        """
        :param system: The name to identify a NETLAB+ system. Defaults to **'default'**.
        :param config: Configuration options. See below for all possible options for the :attr:`config`.
        :param config_path: File path location of a JSON config file to be used instead of the default location.
        """
        self._loop = asyncio.new_event_loop()
        asyncio.set_event_loop(self._loop)

        self._async_client = NetlabClient(system=system, config=config, config_path=config_path)
        self._connection = self._loop.run_until_complete(self._async_client.__aenter__())

    def __repr__(self):
        return str(self._async_client)

    def __dir__(self):
        return list(super().__dir__()) + list(self._async_client._connection.__dir__())

    def __getattr__(self, name):
        from functools import wraps

        func = getattr(self._connection, name)

        @wraps(func)
        def wrapper(*args, **kwargs):
            maybe_co = func(*args, **kwargs)
            if asyncio.iscoroutine(maybe_co):
                result = self._loop.run_until_complete(maybe_co)
            else:
                result = maybe_co
            if hasattr(result, "__anext__"):
                result = _sync_wrap_iter(result, self._loop)
            return result

        return wrapper

    def disconnect(self) -> None:
        """
        Disconnect from NETLAB+.
        """
        self._loop.run_until_complete(self._async_client.__aexit__(None, None, None))
        asyncio.set_event_loop(None)
        self._loop.close()



import logging
import sys
from pathlib import Path
from textwrap import indent

import click
import rich
import tomli
from pydantic import ValidationError
from rich.markup import escape as e
from rich.progress import MofNCompleteColumn, Progress

if sys.version_info >= (3, 11):
    from asyncio import TaskGroup
else:
    from taskgroup import TaskGroup

from netlab.enums import VirtualMachineRole

from ..async_client import NetlabConnection
from ..enums import PodState, RemoveVMS
from . import schema
from .compiler import (
    CompiledDelete,
    CompiledManifest,
    CompiledPod,
    CompiledRange,
    CompiledRangeErrors,
)
from .fetcher import _fetch_hosts, _fetch_pods
from .types import ExistingPod

log = logging.getLogger(__name__)


async def main(manifest_paths: list[Path], auto_confirm: bool) -> None:
    found_errors = False

    manifests = []

    for manifest_path in manifest_paths:
        log.debug("Loading manifest", extra={"path": manifest_path})

        try:
            manifest = schema.Manifest.model_validate(tomli.loads(manifest_path.read_text()))
            manifests.append(manifest)
        except ValidationError as validation_error:
            print(validation_error)
            rich.print("[red bold]Found errors. Cannot continue.")
            return
            # found_errors |= True

    from .compiler import client_factory_from_manifest

    async with client_factory_from_manifest(manifests[0]) as client:
        pods = await _fetch_pods(client, manifests[0])
        hosts = await _fetch_hosts(client)

    compiled_manifest = CompiledManifest.from_schema(manifests[0], manifest_paths[0], hosts, pods)

    for i, r in enumerate(compiled_manifest.ranges):
        print(f"{compiled_manifest.manifest_path}#{i}")

        if isinstance(r, CompiledRange | CompiledDelete):
            for pod_clo in r.pods:
                conflict_kind = pod_clo.pod_op()
                print(indent(str(conflict_kind), "  "))
        else:
            found_errors |= True
            for error_code, error_str in r.errors:
                rich.print(f"  [red]{e(error_code):20} {e(error_str)}")

    if found_errors:
        rich.print("[red bold]Found errors. Cannot continue.")
        return

    if auto_confirm is False:
        click.confirm("Continue?", abort=True)

    await invoke_manifest(compiled_manifest)


async def invoke_manifest(compiled_manifest: CompiledManifest) -> None:
    pods_to_delete: list[ExistingPod] = []
    pods_to_create: list[CompiledPod] = []
    pods_to_updown: list[CompiledPod] = []

    for r in compiled_manifest.ranges:
        assert not isinstance(r, CompiledRangeErrors)
        for pod in r.pods:
            op = pod.pod_op()

            create, delete = op.op()

            if create:
                pods_to_create.append(create)
            if delete:
                pods_to_delete.append(delete)
                #         pods_to_delete.append(existing_pod)
            # match op:
            #     case pod_op.Absent() | pod_op.Superceded() | pod_op.Present():
            #         pass
            #     case pod_op.Rebuild(pod=pod, existing=existing_pod):
            #         pods_to_create.append(pod)
            #         pods_to_delete.append(existing_pod)
            #     case pod_op.Create(pod=pod):
            #         pods_to_create.append(pod)
            #     case pod_op.Delete(existing=existing_pod)
            #         pods_to_delete.append(existing_pod)

            if isinstance(pod.parent, CompiledRange) and pod.parent.get_role() is not VirtualMachineRole.TEMPLATE:
                pods_to_updown.append(pod)

            # if pod.pod_op() in ("REBUILD", "CREATE"):
            #     pods_to_create.append(pod)
            # if (
            #     pod.pod_op() in ("PRESENT", "REBUILD", "CREATE")
            #     and pod.get_role() is not VirtualMachineRole.TEMPLATE
            # ):
            #     pods_to_updown.append(pod)

    with Progress(
        MofNCompleteColumn(),
        *Progress.get_default_columns(),
    ) as progress:
        async with compiled_manifest.client as client:
            delete_progress = progress.add_task("Deleting pods", total=len(pods_to_delete))
            create_progress = progress.add_task("Creating pods", total=len(pods_to_create))
            updown_progress = progress.add_task("Setting pods", total=len(pods_to_updown))

            throt = compiled_manifest.throttle_delete.make()
            async with TaskGroup() as tg:
                for delpod in reversed(pods_to_delete):
                    bucket = await throt.bucket()
                    task = tg.create_task(bucket(pod_deleter(client, delpod)))
                    task.add_done_callback(lambda f: progress.advance(delete_progress, 1))

            throt = compiled_manifest.throttle_create.make()
            async with TaskGroup() as tg:
                for pod in pods_to_create:
                    bucket = await throt.bucket()
                    task = tg.create_task(bucket(pod_creator(client, pod)))
                    task.add_done_callback(lambda f: progress.advance(create_progress, 1))

            throt = compiled_manifest.throttle_updown.make()
            async with TaskGroup() as tg:
                for pod in pods_to_updown:
                    bucket = await throt.bucket()
                    task = tg.create_task(bucket(pod_onliner(client, pod)))
                    task.add_done_callback(lambda f: progress.advance(updown_progress, 1))


async def pod_creator(client: NetlabConnection, pod: CompiledPod) -> None:
    try:
        print(f"cloning pod {pod.pod_id}")
        pc_clone_specs = pod.parent.pc_clone_specs
        await client.pod_clone_task(
            source_pod_id=pod.parent.source_pod.pod_id,
            clone_pod_id=pod.pod_id,
            clone_pod_name=pod.get_name(),
            pc_clone_specs=pc_clone_specs,
        )

        print(f"done cloning pod {pod.pod_id}")
    except Exception as e:
        if sys.version_info >= (3, 11):
            e.add_note(f"pod_id={pod.pod_id}")
        raise


async def pod_deleter(client: NetlabConnection, pod: ExistingPod) -> None:
    try:
        print(f"Removing pod {pod.pod_id}")
        await client.pod_remove_task(
            pod_id=pod.pod_id,
            remove_vms=RemoveVMS.DISK,
        )

        print(f"Done removing pod {pod.pod_id}")
    except Exception as e:
        if sys.version_info >= (3, 11):
            e.add_note(f"pod_id={pod.pod_id}")
        raise


async def pod_onliner(client: NetlabConnection, pod: CompiledPod) -> None:
    try:
        state = PodState.ONLINE if pod.parent.pod_online else PodState.OFFLINE
        print(f"Setting pod {pod.pod_id} state={state}")
        assert pod.parent
        await client.pod_state_change(
            pod_id=pod.pod_id,
            state=state,
        )
    except Exception as e:
        if sys.version_info >= (3, 11):
            e.add_note(f"pod_id={pod.pod_id}")
        raise



import asyncio

from netlab import enums
from netlab.async_client import NetlabClient, NetlabConnection

SOURCE_POD_ID = 9000
POD_NAME_TEMPLATE = "example_pod_{}"
POD_RANGE = range(900, 910)


async def clone_pod(client: NetlabConnection, pod_number: int):
    pod_name = POD_NAME_TEMPLATE.format(pod_number)

    await client.pod_clone_task(
        source_pod_id=SOURCE_POD_ID,
        clone_pod_id=pod_number,
        clone_pod_name=pod_name,
        pc_clone_specs=[
            {
                "source_snapshot": "master",
                "clone_snapshot": "master",
            }
        ],
    )
    await client.pod_state_change(pod_id=pod_number, state=enums.PodState.ONLINE)

    return pod_name


async def main():
    async with NetlabClient() as client:
        for pod_task in asyncio.as_completed([clone_pod(client, i) for i in POD_RANGE]):
            pod_name = await pod_task

            print(f"Cloned pod: {pod_name}")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import json
import sys

from paepy.ChannelDefinition import CustomSensorResult  # type: ignore

from netlab.async_client import NetlabClient


async def main():
    # interpret first command line parameter as json object
    if len(sys.argv) == 1:
        # for command line debugging, use the default system
        sys_args = '{"params": "default"}'
    else:
        sys_args = sys.argv[1]
    prtg = json.loads(sys_args)
    system = prtg.get("params", "default").lower()

    # PRTG runs sensors as a different user. Specify the Administrator user's netlab config file here
    async with NetlabClient(system, config_path="C:/Users/Administrator/.netlab/config.json") as client:
        # query data from netlab api
        response = await client.system_usage_memory()

        total_b = response["ram"]["total_b"]
        free_b = response["ram"]["free_b"]
        free_pct = int((free_b / total_b) * 100)

        # create sensor result
        result = CustomSensorResult("OK")

        # add primary channel for percentage of free memory
        result.add_channel(channel_name="Percent Free Memory", primary_channel=True, unit="Percent", value=free_pct)

        # additional channels for remaining values
        result.add_channel(channel_name="Free Memory", unit="BytesMemory", value=free_b)
        result.add_channel(channel_name="Total Memory", unit="BytesMemory", value=total_b)

        # print sensor result to stdout
        print(result.get_json_result())


if __name__ == "__main__":
    asyncio.run(main())
